# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import itertools
import logging
import os
import re
import shlex
import shutil
import sys
import textwrap
from collections.abc import Sequence
from contextlib import suppress
from pathlib import Path

from gitfourchette.localization import *
from gitfourchette.qt import *
import gitfourchette.exttools.win32ctrlc

_logger = logging.getLogger(__name__)

_placeholderPattern = re.compile(r"\$[_a-zA-Z0-9]+")

_whichPath = None


class ToolCommands:
    FlatpakSandboxedCommandPrefix = "flatpak:"

    @staticmethod
    def splitCommandTokens(command: str) -> list[str]:
        # Treat command as POSIX even on Windows!
        if WINDOWS:
            command = command.replace("\\", "/")
        return shlex.split(command, posix=True)

    @classmethod
    def isFlatpakRunCommand(cls, tokens: Sequence[str]):
        """
        Return the index of the REF token (application ID) in a "flatpak run" command.
        Return 0 if this isn't a valid "flatpak run" command.

        For example, this function would return 4 for "flatpak --verbose run --arch=aarch64 com.example.app"
        because "com.example.app" is token #4.
        """

        i = 0

        try:
            # First token must be flatpak or *bin/flatpak
            if not (tokens[i] == "flatpak" or tokens[i].endswith("bin/flatpak")):
                return 0
            i += 1

            # First positional argument must be `run`
            while tokens[i].startswith("-"):  # Skip switches
                i += 1
            if tokens[i] != "run":
                return 0
            i += 1

            # Get ref token (force IndexError if there's none)
            while tokens[i].startswith("-"):  # Skip switches
                i += 1
            _dummy = tokens[i]
            return i

        except IndexError:
            return 0

    @classmethod
    def replaceProgramTokenInCommand(cls, command: str, *newProgramTokens: str):
        tokens = ToolCommands.splitCommandTokens(command)
        tokens = list(newProgramTokens) + tokens[1:]

        newCommand = shlex.join(tokens)

        # Remove single quotes added around our placeholders by shlex.join()
        # (e.g. '$L' --> $L, '--output=$M' --> $M)
        newCommand = re.sub(r" '(\$[0-9A-Z]+)'", r" \1", newCommand, flags=re.IGNORECASE)
        newCommand = re.sub(r" '(--?[a-z0-9\-_]+=\$[0-9A-Z]+)'", r" \1", newCommand, flags=re.IGNORECASE)

        return newCommand

    @classmethod
    def checkCommand(cls, command: str, *placeholders: str) -> str:
        try:
            cls.compileCommand(command, dict.fromkeys(placeholders, "PLACEHOLDER"), [])
            return ""
        except ValueError as e:
            return str(e)

    @classmethod
    def findPlaceholderTokens(cls, originalTokens: list[str]):
        for token in originalTokens:
            yield from _placeholderPattern.findall(token)

    @classmethod
    def injectReplacements(cls, tokens: list[str], replacements: dict[str, str]) -> list[str]:
        newTokens = []

        mandatoryTokensRemaining = list(replacements.keys())

        for token in tokens:
            matches = list(_placeholderPattern.finditer(token))

            for match in reversed(matches):
                placeholder = match.group()
                try:
                    replacement = replacements[placeholder]
                except KeyError as replacementNotFoundError:
                    raise ValueError(_("Unknown placeholder: {0}", placeholder)
                                     ) from replacementNotFoundError
                with suppress(ValueError):
                    mandatoryTokensRemaining.remove(placeholder)
                token = token[:match.start()] + replacement + token[match.end():]

            if token:
                newTokens.append(token)

        if mandatoryTokensRemaining:
            raise ValueError(_n("Missing placeholder:", "Missing placeholders:", len(mandatoryTokensRemaining))
                             + " " + ", ".join(mandatoryTokensRemaining))

        return newTokens

    @classmethod
    def compileCommand(
            cls,
            command: str,
            replacements: dict[str, str],
            positional: list[str],
            directory: str = "",
            detached: bool = True,
    ) -> tuple[list[str], str]:
        tokens = ToolCommands.splitCommandTokens(command)
        tokens = ToolCommands.injectReplacements(tokens, replacements)
        tokens.extend(positional)  # Append other paths to end of command line

        if tokens and tokens[0].startswith("assets:"):
            internalAssetFile = QFile(tokens[0])
            tokens[0] = internalAssetFile.fileName()

        # Find appropriate workdir
        if not directory:
            for argument in itertools.chain(replacements.values(), positional):
                if not argument:
                    continue
                directory = os.path.dirname(argument)
                if os.path.isdir(directory):
                    break

        # Launch macOS app bundle
        if MACOS and tokens and tokens[0].endswith(".app"):
            flags = "-nF"
            if not detached:
                flags += "W"
            tokens = ["/usr/bin/open", flags, tokens[0], "--args"] + tokens[1:]

        return tokens, directory

    @classmethod
    def setQProcessEnvironment(cls, process: QProcess, environment: dict[str, str] | None):
        if not environment:
            return
        processEnvironment = QProcessEnvironment.systemEnvironment()
        for k, v in environment.items():
            processEnvironment.insert(k, v)
        process.setProcessEnvironment(processEnvironment)

    @classmethod
    def filterQProcessEnvironment(cls, process: QProcess) -> dict[str, str]:
        processEnvironment = process.processEnvironment()
        env = {}
        for key in processEnvironment.keys():  # noqa: SIM118
            value = processEnvironment.value(key)
            if value != INITIAL_ENVIRONMENT.get(key, None):
                env[key] = value
        return env

    @classmethod
    def isFlatpakInstalled(cls, flatpakRef: str) -> bool:
        if not FREEDESKTOP:
            return False
        text = cls.runSync("flatpak", "info", "--show-ref", flatpakRef)
        return bool(text.strip())

    @classmethod
    def wrapFlatpakCommand(cls, process: QProcess, detached=False):
        if not FREEDESKTOP:  # pragma: no cover
            return

        program = process.program()

        if FLATPAK and program.startswith(cls.FlatpakSandboxedCommandPrefix):
            program = program.removeprefix(cls.FlatpakSandboxedCommandPrefix)
            process.setProgram(program)
            return

        tokens = [program] + process.arguments()
        directory = process.workingDirectory()

        # If we're running a "flatpak run" command, add --env, --cwd, and --filesystem.
        flatpakRun = cls.isFlatpakRunCommand(tokens)
        if flatpakRun != 0:
            extra = [f"--env={k}={v}" for k, v in cls.filterQProcessEnvironment(process).items()]

            # Expose directory
            if directory:
                extra += [f"--cwd={directory}", f"--filesystem={directory}"]

            tokens = tokens[:flatpakRun] + extra + tokens[flatpakRun:]

        # If GitFourchette itself is running as a Flatpak, launch the command with "flatpak-spawn".
        if FLATPAK:
            assert program != "flatpak-spawn", "process already flatpak-spawn compatible"

            # Run external tool via flatpak-spawn --host (outside flatpak sandbox).
            extra = ["flatpak-spawn", "--host"]

            # Set --watch-bus if we want the process to die with us.
            # If detached, don't set --watch-bus and DO NOT USE QProcess.startDetached()!
            # (Can't get return code otherwise.)
            if not detached:
                extra += ["--watch-bus"]

            if directory:
                extra += [f"--directory={directory}"]

            # Forward environment variables (only needed if not already done in `flatpak run`)
            if not flatpakRun:
                extra += [f"--env={k}={v}" for k, v in cls.filterQProcessEnvironment(process).items()]

            # Run command through `env` to get return code 127 if the command is missing.
            # (Note that running ANOTHER flatpak via 'flatpak run' won't return 127).
            if not flatpakRun:
                extra += ["/usr/bin/env"]

            extra += ["--"]

            tokens = extra + tokens

        process.setProgram(tokens[0])
        process.setArguments(tokens[1:])

    @classmethod
    def runSync(cls, *args: str, directory: str = "", strict=False) -> str:
        process = QProcess(None)
        process.setProgram(args[0])
        process.setArguments(args[1:])
        if directory:
            process.setWorkingDirectory(directory)
        cls.wrapFlatpakCommand(process)
        _logger.info(f"runSync: {shlex.join([process.program()] + process.arguments())}")
        process.setProcessChannelMode(QProcess.ProcessChannelMode.ForwardedErrorChannel)
        process.start()
        didFinish = process.waitForFinished()

        if strict:
            if not didFinish:
                raise ChildProcessError(f"process did not finish ({process.error()})")
            elif process.exitCode() != 0:
                raise ChildProcessError(f"process exited with code {process.exitCode()}")

        if not didFinish or process.exitCode() != 0:
            return ""

        return process.readAll().data().decode(errors="replace")

    @classmethod
    def which(cls, name: str):
        global _whichPath
        if FLATPAK and _whichPath is None:
            # Cache host's $PATH
            _whichPath = cls.runSync("sh", "-c", "echo $PATH").rstrip()
        return shutil.which(name, path=_whichPath)

    @classmethod
    def makeTerminalScript(cls, workdir: str, commandTokens: list[str]) -> str:
        # While we could simply export the parameters as environment variables
        # then run termcmd.sh from the static assets, this approach fails if
        # the terminal is a Flatpak (custom env vars always seem to be erased).
        # That's why we generate a bespoke launcher script on the fly
        # (which sources termcmd.sh).

        # When running as a Flatpak, the assets folder lives in an app-specific
        # mount point. I don't think we can get its "real" path on the host.
        # So, copy termcmd.sh out of the assets folder before passing it to the
        # external terminal program.
        termcmdPath = Path(qTempDir()) / "termcmd.sh"
        if not termcmdPath.exists():
            termcmdSourcePath = QFile("assets:termcmd.sh").fileName()
            shutil.copyfile(termcmdSourcePath, termcmdPath)

        yKey = "y"
        nKey = "n"
        exitMessage = _("Command exited with code:")
        keyPrompt = _("Continue in a shell?") + f" [{yKey.upper()}/{nKey}]"

        script = textwrap.dedent(f"""\
            #!/usr/bin/env bash
            _GF_COMMAND={shlex.quote(shlex.join(commandTokens))}
            _GF_APPNAME={shlex.quote(qAppName())}
            _GF_YKEY={shlex.quote(yKey)}
            _GF_NKEY={shlex.quote(nKey)}
            _GF_EXITMESSAGE={shlex.quote(exitMessage)}
            _GF_KEYPROMPT={shlex.quote(keyPrompt)}
            _GF_WORKDIR={shlex.quote(workdir)}
            _GF_TESTMODE={"1" if APP_TESTMODE else "0"}
            source {shlex.quote(str(termcmdPath))}
        """)

        cacheKey = f"{hash(script):x}"
        path = Path(qTempDir()) / f"terminal_{cacheKey}.sh"
        if not path.exists():
            path.write_text(script, "utf-8")
            path.chmod(0o755)
        return str(path)

    @classmethod
    def spawnNewInstance(cls, module: str, sandbox: bool = True) -> list[str]:
        """
        Prepare a command to spawn another instance of this application.
        """
        assert module.startswith("gitfourchette"), module

        if FLATPAK and not sandbox:
            # Command to be called by host system's git, which lives outside the
            # flatpak sandbox. Spawn another instance of our flatpak.
            tokens = ["flatpak", "run", "--command=python", FLATPAK_ID, "-m", module]
        elif "APPIMAGE" in INITIAL_ENVIRONMENT:
            # AVOID "sys.executable" (path to .AppImage file) to spawn a child,
            # may cause SIGTTIN in parent process!
            python = sys.orig_argv[0]  # python in AppImage tmpfs mount
            tokens = [python, "-m", module]
        elif PYINSTALLER_MEIPASS:
            tokens = [sys.executable, f"--start-module={module}"]
        else:
            # Just spawn another Python interpreter.
            tokens = [sys.executable, "-m", module]

        # The new instance is supposed to be a child process, so it will inherit
        # the environment variables of the current process. We don't need to
        # explicitly copy them to 'env'. For reference, some of the relevant
        # variables for a successful child process include:
        # - PYTHONPATH (":".join(sys.path))
        # - For AppImage only: APPIMAGE, APPDIR, ARGV0, SSL_CERT_FILE
        # Without these, askpass.sh may fail to start unless it's a child
        # process of a GitFourchette instance.

        assert tokens
        return tokens

    @classmethod
    def terminatePlus(cls, process: QProcess | None):
        """
        QProcess.terminate() plus Windows console-mode quirks
        """
        if process is None:
            return

        if process.state() == QProcess.ProcessState.NotRunning:
            return

        _logger.info(f"Terminating process {process.program()} (PID {process.processId()})")
        process.terminate()

        if WINDOWS:
            tokens = ToolCommands.spawnNewInstance(gitfourchette.exttools.win32ctrlc.__name__)
            tokens.append(str(process.processId()))

            ctrlC = QProcess(None)
            ctrlC.setObjectName("win32ctrlc")
            ctrlC.setProgram(tokens[0])
            ctrlC.setArguments(tokens[1:])
            ctrlC.startDetached()



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations  # TODO: Remove once we can drop support for Python <= 3.13

from gitfourchette.appconsts import *
from gitfourchette.search.searchprovider import SearchProvider

if TYPE_CHECKING:
    from gitfourchette.sidebar.sidebar import Sidebar


class SidebarSearch(SearchProvider):
    sidebar: Sidebar

    def __init__(self, sidebar: Sidebar):
        super().__init__(sidebar)
        self.sidebar = sidebar

    def invalidate(self):
        super().invalidate()

        # Wipe term from filter, return to permanent collapse state
        self.setTerm("")

        # After returning to the permanent collapse state, force expand the node
        # that the user selected while the filter was active
        node = self.sidebar.selectedNode()
        if node is not None and not self.sidebar.sidebarModel.isAncestryChainExpanded(node):
            self.sidebar.selectionModel().clear()
            self.sidebar.selectNode(node)

    def _termChanged(self):
        term = self.term()
        self.sidebar.filterModel.setFilterText(term)

        if term:
            self.sidebar.setCollapseStateLayer(transient=True)
            self.sidebar.expandAll()
        else:
            self.sidebar.setCollapseStateLayer(transient=False)



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import warnings
from collections.abc import Callable, Iterable
from contextlib import suppress

from gitfourchette import trtables
from gitfourchette import porcelain
from gitfourchette import settings
from gitfourchette.application import GFApplication
from gitfourchette.exttools.usercommand import UserCommand
from gitfourchette.forms.searchbar import SearchBar
from gitfourchette.localization import *
from gitfourchette.nav import NavLocator
from gitfourchette.porcelain import Oid, RefPrefix, Repo
from gitfourchette.qt import *
from gitfourchette.repomodel import RepoModel, UC_FAKEREF
from gitfourchette.repoprefs import RefSort
from gitfourchette.sidebar.sidebardelegate import SidebarDelegate, SidebarClickZone
from gitfourchette.sidebar.sidebarfilter import SidebarFilter
from gitfourchette.sidebar.sidebarmodel import SidebarModel, SidebarNode, SidebarItem
from gitfourchette.sidebar.sidebarsearch import SidebarSearch
from gitfourchette.tasks import *
from gitfourchette.toolbox import *
from gitfourchette.webhost import WebHost

INVALID_MOUSEPRESS = (-1, SidebarClickZone.Invalid)


class Sidebar(QTreeView):
    toggleHideRefPattern = Signal(str, bool)
    openSubmoduleRepo = Signal(str)
    openSubmoduleFolder = Signal(str)
    statusMessage = Signal(str)

    sidebarModel: SidebarModel
    filterModel: SidebarFilter
    selectionBackup: SidebarNode | None
    mousePressCache: tuple[int, SidebarClickZone]

    def __init__(self, parent):
        super().__init__(parent)
        self.repoWidget = parent

        self.setObjectName("Sidebar")
        self.setMouseTracking(True)  # for eye icons
        self.setMinimumWidth(128)
        self.setIndentation(16)
        self.setHeaderHidden(True)
        self.setUniformRowHeights(True)  # large sidebars update twice as fast with this, but we can't have thinner spacers
        self.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)

        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self.onCustomContextMenuRequested)

        self.sidebarModel = SidebarModel(self)
        self.filterModel = SidebarFilter(self)
        self.filterModel.setSourceModel(self.sidebarModel)
        self.setModel(self.filterModel)

        self.toggleHideRefPattern.connect(self.sidebarModel.clearCachedTooltip)

        self.setItemDelegate(SidebarDelegate(self))

        self.expanded.connect(self.onIndexExpanded)
        self.collapsed.connect(self.onIndexCollapsed)
        self.selectionBackup = None

        # When clicking on a row, information about the clicked row and "zone"
        # is kept until mouseReleaseEvent.
        self.mousePressCache = INVALID_MOUSEPRESS

        self.searchBar = SearchBar(self, SidebarSearch(self))
        self.searchBar.ui.backwardButton.hide()
        self.searchBar.ui.forwardButton.hide()
        self.searchBar.hide()

        GFApplication.instance().prefsChanged.connect(self.refreshPrefs)
        self.refreshPrefs()

        makeWidgetShortcut(self, self.onReturnShortcut, "Return", "Enter")
        makeWidgetShortcut(self, self.onDeleteShortcut, "Delete")
        makeWidgetShortcut(self, self.onRenameShortcut, "F2")
        makeWidgetShortcut(self, self.searchBar.hideOrBeep, "Escape")

    def filterIndexToNode(self, index: QModelIndex) -> SidebarNode:
        assert index.isValid()
        assert index.model() is self.filterModel
        index = self.filterModel.mapToSource(index)
        return SidebarNode.fromIndex(index)

    def nodeToFilterIndex(self, node: SidebarNode) -> QModelIndex:
        index = self.sidebarModel.createIndexFromNode(node)
        assert index.isValid()
        assert index.model() is self.sidebarModel
        return self.filterModel.mapFromSource(index)

    def drawBranches(self, painter, rect, index):
        """
        (overridden function)
        Prevent drawing the default "branches" and expand/collapse indicators.
        We draw the expand/collapse indicator ourselves in SidebarDelegate.
        """
        # Overriding this function has the same effect as adding this line in the QSS:
        # Sidebar::branch { image: none; border-image: none; }
        return

    def visualRect(self, index: QModelIndex) -> QRect:
        """
        Required so the theme can properly draw unindented rows.
        """

        vr = super().visualRect(index)

        if index.isValid():
            node = self.filterIndexToNode(index)
            SidebarDelegate.unindentRect(node.kind, vr, self.indentation())

        return vr

    def refSortMenu(self, prefKey: str) -> list[ActionDef]:
        repoModel = self.sidebarModel.repoModel
        repoPrefs = repoModel.prefs

        defaultMode = settings.prefs.refSort
        currentMode = getattr(repoPrefs, prefKey)
        assert isinstance(currentMode, RefSort)
        if currentMode == RefSort.UseGlobalPref:
            currentMode = defaultMode

        def setSortMode(newMode: RefSort):
            if currentMode == newMode:
                return
            setattr(repoPrefs, prefKey, newMode)
            repoPrefs.refSortClearTimestamp = QDateTime.currentSecsSinceEpoch()
            repoPrefs.setDirty()
            self.backUpSelection()
            self.refresh(repoModel)
            self.restoreSelectionBackup()

        submenu = []
        for sortMode in RefSort:
            if sortMode == RefSort.UseGlobalPref:
                continue
            caption = trtables.enum(sortMode)
            action = ActionDef(
                caption,
                lambda m=sortMode: setSortMode(m),
                radioGroup=f"sortBy-{prefKey}",
                checkState=1 if currentMode == sortMode else -1)
            submenu.append(action)

        return submenu

    def makeNodeMenu(self, node: SidebarNode):
        actions = []

        model = self.sidebarModel
        repo = model.repo
        item = node.kind
        data = node.data
        isExplicitlyShown = model.isExplicitlyShown(node)
        isExplicitlyHidden = model.isExplicitlyHidden(node)
        mainWindow = GFApplication.instance().mainWindow

        if item == SidebarItem.WorkdirHeader:
            actions += self.repoWidget.contextMenuItems()

        elif item == SidebarItem.UncommittedChanges:
            actions += [
                TaskBook.action(self, NewCommit, accel="C"),
                TaskBook.action(self, AmendCommit, accel="A"),
                ActionDef.SEPARATOR,
                TaskBook.action(self, NewStash, accel="S"),
                TaskBook.action(self, ExportWorkdirAsPatch, accel="X"),
                *mainWindow.contextualUserCommands(UserCommand.Token.Workdir),
            ]

        elif item == SidebarItem.LocalBranchesHeader:
            actions += [
                TaskBook.action(self, NewBranchFromHead, _("&New Branch…")),
                ActionDef.SEPARATOR,
                ActionDef(_("Sort By"), submenu=self.refSortMenu("sortBranches")),
            ]

            if any(n.kind == SidebarItem.RefFolder for n in node.children):
                actions += [
                    ActionDef(_("Collapse All Folders"), lambda: self.collapseChildFolders(node)),
                    ActionDef(_("Expand All Folders"), lambda: self.expandChildFolders(node)),
                ]

        elif item == SidebarItem.LocalBranch:
            refName = data
            prefix, branchName = RefPrefix.split(data)
            assert prefix == RefPrefix.HEADS
            branch = repo.branches.local[branchName]

            activeBranchName = model.repoModel.homeBranch
            isCurrentBranch = branch and branch.is_checked_out()
            hasUpstream = bool(branch.upstream)
            upstreamBranchName = "" if not hasUpstream else branch.upstream.shorthand
            upstreamSubmenuTitle = _("&Upstream Branch")

            upstreamMissing = False
            if not upstreamBranchName:
                with suppress(KeyError):
                    upstreamBranchName = self.sidebarModel.repoModel.upstreams[branchName]
                    hasUpstream = True
                    upstreamMissing = True
                    upstreamSubmenuTitle += " " +_("(Missing)")

            thisBranchDisplay = lquoe(branchName)
            activeBranchDisplay = lquoe(activeBranchName)
            upstreamBranchDisplay = lquoe(upstreamBranchName)

            userCommandTokens = [UserCommand.Token.Commit, UserCommand.Token.Ref]
            if isCurrentBranch:
                userCommandTokens.extend([
                    UserCommand.Token.Head,
                    UserCommand.Token.HeadBranch,
                    UserCommand.Token.HeadUpstream
                ])

            actions += [
                TaskBook.action(
                    self,
                    SwitchBranch,
                    _("&Switch to {0}…", thisBranchDisplay),
                    taskArgs=branchName,
                ).replace(enabled=not isCurrentBranch),

                TaskBook.action(
                    self,
                    MergeBranch,
                    _("&Merge into {0}…", activeBranchDisplay),
                    taskArgs=refName,
                ).replace(enabled=not isCurrentBranch and activeBranchName),

                ActionDef.SEPARATOR,

                TaskBook.action(
                    self,
                    FetchRemoteBranch,
                    _("&Fetch {0}", upstreamBranchDisplay) if hasUpstream else _("Fetch"),
                    taskArgs=upstreamBranchName if hasUpstream else None,
                ).replace(enabled=hasUpstream),

                TaskBook.action(
                    self,
                    PullBranch,
                    _("Pu&ll from {0}…", upstreamBranchDisplay) if hasUpstream else _("Pull…"),
                ).replace(enabled=hasUpstream and isCurrentBranch),

                TaskBook.action(
                    self,
                    FastForwardBranch,
                    _("Fast-Forward to {0}…", upstreamBranchDisplay) if hasUpstream else _("Fast-Forward…"),
                    taskArgs=branchName,
                ).replace(enabled=hasUpstream and not upstreamMissing),

                TaskBook.action(
                    self,
                    PushBranch,
                    _("&Push to {0}…", upstreamBranchDisplay) if hasUpstream else _("Push…"),
                    taskArgs=branchName,
                ),

                ActionDef(
                    upstreamSubmenuTitle,
                    submenu=self.makeUpstreamSubmenu(repo, branchName, upstreamBranchName, upstreamMissing),
                    icon="git-upstream-missing" if upstreamMissing else ""
                ),

                ActionDef.SEPARATOR,

                TaskBook.action(self, RenameBranch, _("Re&name…"), taskArgs=branchName),

                TaskBook.action(self, DeleteBranch, _("&Delete…"), taskArgs=branchName),

                ActionDef.SEPARATOR,

                TaskBook.action(self, NewBranchFromRef, _("New &Branch Here…"), taskArgs=refName),

                ActionDef.SEPARATOR,

                ActionDef(_("&Hide in Graph"),
                          lambda: self.wantHideNode(node),
                          checkState=[-1, 1][isExplicitlyHidden],
                          tip=_("Hide this branch from the graph (effective if no other branches/tags point here)"),
                          icon="view-hidden"),

                ActionDef(_("Hide &All But This"),
                          lambda: self.wantHideNode(node, True),
                          checkState=[-1, 1][isExplicitlyShown],
                          icon="view-exclusive"),

                *mainWindow.contextualUserCommands(*userCommandTokens),
            ]

        elif item == SidebarItem.DetachedHead:
            actions += [TaskBook.action(self, NewBranchFromHead, _("New &Branch Here…")), ]

        elif item == SidebarItem.RemoteBranch:
            activeBranchName = model.repoModel.homeBranch
            activeBranchDisplay = lquoe(activeBranchName) if activeBranchName else _("(no current branch)")

            refName = data
            prefix, shorthand = RefPrefix.split(refName)
            assert prefix == RefPrefix.REMOTES

            remoteName, remoteBranchName = porcelain.split_remote_branch_shorthand(shorthand)
            remoteUrl = self.sidebarModel.repo.remotes[remoteName].url
            webUrl, webHost = WebHost.makeLink(remoteUrl, remoteBranchName)
            webActions = []
            if webUrl:
                webActions = [
                    ActionDef(
                        _("Visit Web Page on {0}", escamp(webHost)),
                        lambda: QDesktopServices.openUrl(QUrl(webUrl)),
                        icon="internet-web-browser",
                        tip=f"<p style='white-space: pre'>{escape(webUrl)}</p>",
                    ),
                    ActionDef.SEPARATOR,
                ]

            actions += [
                TaskBook.action(self, NewBranchFromRef, _("New Local &Branch Here…"), taskArgs=refName),

                TaskBook.action(self, FetchRemoteBranch, _("&Fetch New Commits"), taskArgs=shorthand),

                ActionDef.SEPARATOR,

                TaskBook.action(self, MergeBranch, _("&Merge into {0}…", activeBranchDisplay),
                                taskArgs=refName, enabled=bool(activeBranchName)),

                ActionDef.SEPARATOR,

                TaskBook.action(self, RenameRemoteBranch, taskArgs=shorthand),

                TaskBook.action(self, DeleteRemoteBranch, taskArgs=shorthand),

                ActionDef.SEPARATOR,

                *webActions,

                ActionDef(_("&Hide in Graph"),
                          lambda: self.wantHideNode(node),
                          checkState=[-1, 1][isExplicitlyHidden],
                          icon="view-hidden"),

                ActionDef(_("Hide &All But This"),
                          lambda: self.wantHideNode(node, True),
                          checkState=[-1, 1][isExplicitlyShown],
                          icon="view-exclusive"),

                *mainWindow.contextualUserCommands(UserCommand.Token.Commit, UserCommand.Token.Ref, UserCommand.Token.Remote),
            ]

        elif item == SidebarItem.Remote:
            remoteUrl = model.repo.remotes[data].url
            webUrl, webHost = WebHost.makeLink(remoteUrl)

            collapseActions = []
            if any(n.kind == SidebarItem.RefFolder for n in node.children):
                collapseActions = [
                    ActionDef(_("Collapse All Folders"), lambda: self.collapseChildFolders(node)),
                    ActionDef(_("Expand All Folders"), lambda: self.expandChildFolders(node)),
                ]

            webActions = []
            if webUrl:
                webActions = [
                    ActionDef(
                        _("Visit Web Page on {0}", escamp(webHost)),
                        lambda: QDesktopServices.openUrl(QUrl(webUrl)),
                        icon="internet-web-browser",
                        tip=f"<p style='white-space: pre'>{escape(webUrl)}</p>",
                    ),
                ]

            actions += [
                TaskBook.action(self, EditRemote, accel="E", taskArgs=data),

                TaskBook.action(self, FetchRemotes, accel="F", taskArgs=data),

                ActionDef.SEPARATOR,

                TaskBook.action(self, DeleteRemote, accel="R", taskArgs=data),

                ActionDef.SEPARATOR,

                *webActions,

                ActionDef(_("Copy Remote &URL"),
                          lambda: self.copyToClipboard(remoteUrl)),

                ActionDef.SEPARATOR,

                *collapseActions,

                ActionDef(_("&Hide in Graph"),
                          lambda: self.wantHideNode(node),
                          checkState=[-1, 1][isExplicitlyHidden],
                          icon="view-hidden"),

                ActionDef(_("Hide &All But This"),
                          lambda: self.wantHideNode(node, True),
                          checkState=[-1, 1][isExplicitlyShown],
                          icon="view-exclusive"),

                *mainWindow.contextualUserCommands(UserCommand.Token.Remote),
            ]

        elif item == SidebarItem.RemotesHeader:
            actions += [
                TaskBook.action(self, NewRemote, accel="A"),
                TaskBook.action(self, FetchRemotes, accel="F"),
                ActionDef.SEPARATOR,
                ActionDef(_("Sort Remote Branches By"), submenu=self.refSortMenu("sortRemoteBranches")),
            ]

        elif item == SidebarItem.RefFolder:
            if node.data.startswith(RefPrefix.HEADS):
                actions += [
                    ActionDef(_("Re&name Folder…"), lambda: self.wantRenameNode(node)),
                    ActionDef(_("&Delete Folder…"), lambda: self.wantDeleteNode(node)),
                    ActionDef.SEPARATOR,
                ]

            actions += [
                ActionDef(_("&Hide in Graph"),
                          lambda: self.wantHideNode(node),
                          checkState=[-1, 1][isExplicitlyHidden],
                          icon="view-hidden"),

                ActionDef(_("Hide &All But This"),
                          lambda: self.wantHideNode(node, True),
                          checkState=[-1, 1][isExplicitlyShown],
                          icon="view-exclusive"),
            ]

        elif item == SidebarItem.StashesHeader:
            actions += [
                TaskBook.action(self, NewStash, accel="S"),
            ]

        elif item == SidebarItem.Stash:
            oid = Oid(hex=data)

            actions += [
                TaskBook.action(self, ApplyStash, accel="A", taskArgs=oid),

                TaskBook.action(self, ExportStashAsPatch, accel="X", taskArgs=oid),

                ActionDef.SEPARATOR,

                TaskBook.action(self, DropStash, accel="D", taskArgs=oid),

                ActionDef.SEPARATOR,

                ActionDef(_("Reveal &Parent Commit"), lambda: self.revealStashParent(oid)),
            ]

        elif item == SidebarItem.TagsHeader:
            refspecs = [ref for ref in self.sidebarModel.repoModel.refs
                        if ref.startswith(RefPrefix.TAGS)]

            actions += [
                TaskBook.action(self, NewTag, _("&New Tag on HEAD Commit…")),
                ActionDef(_("&Push All Tags To"), submenu=self.pushRefspecMenu(refspecs), enabled=bool(refspecs)),
                ActionDef.SEPARATOR,
                ActionDef(_("Sort By"), submenu=self.refSortMenu("sortTags")),
            ]

        elif item == SidebarItem.Tag:
            prefix, shorthand = RefPrefix.split(data)
            assert prefix == RefPrefix.TAGS
            refspecs = [data]

            target = self.sidebarModel.repoModel.refs[data]

            actions += [
                TaskBook.action(self, CheckoutCommit, _("&Check Out Tagged Commit…"), taskArgs=target),
                TaskBook.action(self, DeleteTag, _("&Delete Tag…"), taskArgs=shorthand),
                ActionDef.SEPARATOR,
                ActionDef(_("Push To"), submenu=self.pushRefspecMenu(refspecs)),
            ]

        elif item == SidebarItem.SubmodulesHeader:
            submodules = self.sidebarModel.repoModel.submodules

            actions += [
                TaskBook.action(self, UpdateSubmodulesRecursive, enabled=bool(submodules)),
            ]

        elif item == SidebarItem.Submodule:
            model = self.sidebarModel
            repo = model.repo

            actions += [
                ActionDef(_("&Open Submodule in New Tab"),
                          lambda: self.openSubmoduleRepo.emit(data)),

                ActionDef(_("Open Submodule &Folder"),
                          lambda: self.openSubmoduleFolder.emit(data)),

                ActionDef(_("Copy &Path"),
                          lambda: self.copyToClipboard(repo.in_workdir(data))),

                ActionDef.SEPARATOR,

                TaskBook.action(self, UpdateSubmodule, taskArgs=data),

                ActionDef.SEPARATOR,

                TaskBook.action(self, RemoveSubmodule, taskArgs=data),
            ]

        # --------------------

        if not actions:
            return None

        menu = ActionDef.makeQMenu(self, actions)
        menu.setToolTipsVisible(True)
        return menu

    def onCustomContextMenuRequested(self, point: QPoint):
        if APP_TESTMODE and point == QPoint_zero:
            # Unit testing hack: spawn context menu on selected item
            index = self.selectedIndexes()[0]
        else:
            # Real world: don't spawn context menu in blank areas/separators
            index = self.indexAt(point)
            if not index.isValid():
                return

        node = self.filterIndexToNode(index)
        menu = self.makeNodeMenu(node)

        if menu:
            menu.aboutToHide.connect(menu.deleteLater)
            menu.popup(self.mapToGlobal(point))

    def refresh(self, repoModel: RepoModel):
        self.sidebarModel.rebuild(repoModel)
        self.restoreExpandedItems()

    def backUpSelection(self):
        self.selectionBackup = self.selectedNode()

    def clearSelectionBackup(self):
        self.selectionBackup = None

    def restoreSelectionBackup(self):
        if self.selectionBackup is None:
            return

        try:
            node = self.findNode(self.selectionBackup.isSimilarEnoughTo)
        except StopIteration:
            pass
        else:
            with QSignalBlockerContext(self):
                self.selectNode(node)

        self.clearSelectionBackup()

    def refreshPrefs(self):
        self.setVerticalScrollMode(settings.prefs.listViewScrollMode)
        self.setAnimated(settings.prefs.animations)

    def repaintUncommittedChanges(self):
        index = self.indexForRef(UC_FAKEREF)
        self.update(index)

    def wantSelectNode(self, node: SidebarNode):
        if self.signalsBlocked():  # Don't bother with the jump if our signals are blocked
            return

        assert isinstance(node, SidebarNode)

        item = node.kind

        if item == SidebarItem.UncommittedChanges:
            locator = NavLocator.inWorkdir()
        elif item == SidebarItem.UnbornHead:
            locator = NavLocator.inWorkdir()
        elif item == SidebarItem.DetachedHead:
            locator = NavLocator.inRef("HEAD")
        elif item == SidebarItem.LocalBranch:
            locator = NavLocator.inRef(node.data)
        elif item == SidebarItem.RemoteBranch:
            locator = NavLocator.inRef(node.data)
        elif item == SidebarItem.Tag:
            locator = NavLocator.inRef(node.data)
        elif item == SidebarItem.Stash:
            locator = NavLocator.inCommit(Oid(hex=node.data))
        else:
            return

        Jump.invoke(self, locator)

    def wantEnterNode(self, node: SidebarNode | None):
        if node is None:
            QApplication.beep()
            return
        assert isinstance(node, SidebarNode)

        item = node.kind

        if item == SidebarItem.Spacer:
            pass

        elif item == SidebarItem.LocalBranch:
            SwitchBranch.invoke(self, node.data.removeprefix(RefPrefix.HEADS))

        elif item == SidebarItem.Remote:
            EditRemote.invoke(self, node.data)

        elif item == SidebarItem.RemotesHeader:
            NewRemote.invoke(self)

        elif item == SidebarItem.LocalBranchesHeader:
            NewBranchFromHead.invoke(self)

        elif item == SidebarItem.UncommittedChanges:
            NewCommit.invoke(self)

        elif item == SidebarItem.Submodule:
            self.openSubmoduleRepo.emit(node.data)

        elif item == SidebarItem.StashesHeader:
            NewStash.invoke(self)

        elif item == SidebarItem.Stash:
            oid = Oid(hex=node.data)
            ApplyStash.invoke(self, oid)

        elif item == SidebarItem.RemoteBranch:
            NewBranchFromRef.invoke(self, node.data)

        elif item == SidebarItem.DetachedHead:
            NewBranchFromHead.invoke(self)

        elif item == SidebarItem.TagsHeader:
            NewTag.invoke(self)

        elif item == SidebarItem.Tag:
            target = self.sidebarModel.repoModel.refs[node.data]
            CheckoutCommit.invoke(self, target)

        else:
            QApplication.beep()

    def wantDeleteNode(self, node: SidebarNode | None):
        if node is None:
            QApplication.beep()
            return

        item = node.kind
        data = node.data

        if item == SidebarItem.Spacer:
            pass

        elif item == SidebarItem.LocalBranch:
            assert data.startswith(RefPrefix.HEADS)
            DeleteBranch.invoke(self, data.removeprefix(RefPrefix.HEADS))

        elif item == SidebarItem.Remote:
            DeleteRemote.invoke(self, data)

        elif item == SidebarItem.Stash:
            oid = Oid(hex=data)
            DropStash.invoke(self, oid)

        elif item == SidebarItem.RemoteBranch:
            assert data.startswith(RefPrefix.REMOTES)
            DeleteRemoteBranch.invoke(self, data.removeprefix(RefPrefix.REMOTES))

        elif item == SidebarItem.Tag:
            assert data.startswith(RefPrefix.TAGS)
            DeleteTag.invoke(self, data.removeprefix(RefPrefix.TAGS))

        elif item == SidebarItem.RefFolder:
            prefix, _name = RefPrefix.split(data)
            if prefix == RefPrefix.HEADS:
                DeleteBranchFolder.invoke(self, data)
            else:
                QApplication.beep()

        else:
            QApplication.beep()

    def wantRenameNode(self, node: SidebarNode | None):
        if node is None:
            QApplication.beep()
            return

        item = node.kind
        data = node.data

        if item == SidebarItem.Spacer:
            pass

        elif item == SidebarItem.LocalBranch:
            prefix, name = RefPrefix.split(data)
            assert prefix == RefPrefix.HEADS
            RenameBranch.invoke(self, name)

        elif item == SidebarItem.Remote:
            EditRemote.invoke(self, data)

        elif item == SidebarItem.RemoteBranch:
            RenameRemoteBranch.invoke(self, data)

        elif item == SidebarItem.RefFolder:
            prefix, name = RefPrefix.split(data)
            if prefix == RefPrefix.HEADS:
                RenameBranchFolder.invoke(self, data)
            else:
                QApplication.beep()

        else:
            QApplication.beep()

    def wantHideNode(self, node: SidebarNode, allButThis: bool = False):
        pattern = node.refMatchingPattern()
        if not pattern:
            return
        self.toggleHideRefPattern.emit(pattern, allButThis)
        self.repaint()

    def selectedNode(self) -> SidebarNode | None:
        try:
            index = self.selectedIndexes()[0]
        except IndexError:
            return None
        else:
            return self.filterIndexToNode(index)

    def onReturnShortcut(self):
        self.wantEnterNode(self.selectedNode())

    def onDeleteShortcut(self):
        self.wantDeleteNode(self.selectedNode())

    def onRenameShortcut(self):
        self.wantRenameNode(self.selectedNode())

    def selectionChanged(self, selected: QItemSelection, deselected: QItemSelection):
        super().selectionChanged(selected, deselected)

        if selected.count() == 0:
            return

        current = selected.indexes()[0]
        if not current.isValid():
            return

        node = self.filterIndexToNode(current)
        self.wantSelectNode(node)

    def resolveClick(self, pos: QPoint) -> tuple[QModelIndex, SidebarNode | None, SidebarClickZone]:
        index = self.indexAt(pos)
        if index.isValid():
            rect = self.visualRect(index)
            node = self.filterIndexToNode(index)
            zone = SidebarDelegate.getClickZone(node, rect, pos.x())
        else:
            node = None
            zone = SidebarClickZone.Invalid
        return index, node, zone

    def mouseMoveEvent(self, event):
        """
        Bypass QTreeView's mouseMoveEvent to prevent the original branch
        expand indicator from taking over mouse input.
        """
        QAbstractItemView.mouseMoveEvent(self, event)

    def mousePressEvent(self, event: QMouseEvent):
        pos = event.position().toPoint()
        index, _node, zone = self.resolveClick(pos)

        # Save click info for mouseReleaseEvent
        self.mousePressCache = (index.row(), zone)

        if zone == SidebarClickZone.Invalid:
            super().mousePressEvent(event)
            return

        if zone == SidebarClickZone.Select:
            self.setCurrentIndex(index)

        event.accept()

    def mouseReleaseEvent(self, event: QMouseEvent):
        pos = event.position().toPoint()
        index, node, zone = self.resolveClick(pos)

        # Only perform special actions (hide, expand) if mouse was released
        # on same row/zone as mousePressEvent
        match = (index.row(), zone) == self.mousePressCache

        # Clear mouse press info on release
        self.mousePressCache = INVALID_MOUSEPRESS

        if (not match
                or node is None
                or zone in [SidebarClickZone.Invalid, SidebarClickZone.Select]):
            super().mouseReleaseEvent(event)
            return

        if zone == SidebarClickZone.Hide:
            allButThis = event.button() == Qt.MouseButton.MiddleButton or self.sidebarModel.isHideAllButThisMode()
            self.wantHideNode(node, allButThis)
            event.accept()
        elif zone == SidebarClickZone.Expand:
            # Toggle expanded state
            if node.mayHaveChildren():
                if self.isExpanded(index):
                    self.collapse(index)
                else:
                    self.expand(index)
            event.accept()
        else:
            warnings.warn(f"Unknown click zone {zone}")  # type: ignore[unreachable]

    def mouseDoubleClickEvent(self, event: QMouseEvent):
        # NOT calling "super().mouseDoubleClickEvent(event)" on purpose.

        pos = event.position().toPoint()
        _index, node, zone = self.resolveClick(pos)

        # Let user collapse/expand/hide a single node in quick succession
        # without triggering a double click
        if zone not in [SidebarClickZone.Invalid, SidebarClickZone.Select]:
            self.mousePressEvent(event)
            return

        if event.button() == Qt.MouseButton.LeftButton and node is not None:
            self.wantEnterNode(node)

    def revealStashParent(self, oid: Oid):
        commit = self.sidebarModel.repo.peel_commit(oid)
        parent = commit.parent_ids[0]
        Jump.invoke(self, NavLocator.inCommit(parent))

    # -------------------------------------------------------------------------
    # Sidebar nodes

    def walk(self):
        return self.sidebarModel.rootNode.walk()

    def findNode(self, predicate: Callable[[SidebarNode], bool]) -> SidebarNode:
        return next(node for node in self.walk() if predicate(node))

    def findNodeByRef(self, ref: str) -> SidebarNode:
        return self.sidebarModel.nodesByRef[ref]

    def findNodesByKind(self, kind: SidebarItem) -> list[SidebarNode]:
        """ Unit test helper """
        return [node for node in self.walk() if node.kind == kind]

    def findNodeByKind(self, kind: SidebarItem) -> SidebarNode:
        """ Unit test helper """
        nodes = self.findNodesByKind(kind)
        if len(nodes) == 0:
            raise KeyError(str(kind))
        if len(nodes) != 1:
            raise ValueError(f"{kind} is not unique")
        return nodes[0]

    def countNodesByKind(self, kind: SidebarItem):
        """ Unit test helper """
        return len(self.findNodesByKind(kind))

    def indexForRef(self, ref: str) -> QModelIndex:
        """
        Return an index in the filtered model for the given ref.
        The returned index may be invalid if the ref is filtered out.
        """
        try:
            node = self.sidebarModel.nodesByRef[ref]
        except KeyError:
            return QModelIndex()  # Invalid index
        else:
            return self.nodeToFilterIndex(node)

    def selectNode(self, node: SidebarNode) -> QModelIndex:
        index = self.nodeToFilterIndex(node)
        self.setCurrentIndex(index)
        return index

    def selectAnyRef(self, *refs: str) -> QModelIndex | None:
        # Early out if any candidate ref is already selected
        with suppress(IndexError):
            index = self.selectedIndexes()[0]
            if index.data(SidebarModel.Role.Ref) in refs:
                return index

        # If several refs point to the same commit, attempt to select
        # the checked-out branch first (or its upstream, if any)
        model = self.sidebarModel
        refCandidates: Iterable[str]
        if model._checkedOut:
            favorRefs = [RefPrefix.HEADS + model._checkedOut,
                         RefPrefix.REMOTES + model._checkedOutUpstream]
            refCandidates = sorted(refs, key=favorRefs.__contains__, reverse=True)
        else:
            refCandidates = refs

        # Find a visible index that matches any of the candidates
        for ref in refCandidates:
            index = self.indexForRef(ref)
            if not index.isValid():
                continue
            node = self.filterIndexToNode(index)
            # Don't force-expand any collapsed indexes
            if model.isAncestryChainExpanded(node):
                self.setCurrentIndex(index)
                return index

        # There are no indices that match the candidates, so select nothing
        self.clearSelection()
        return None

    # -------------------------------------------------------------------------
    # Collapse/expand

    @benchmark
    def restoreExpandedItems(self):
        """
        Assumes all collapsible rows are collapsed before this is called.
        (After rebuilding the model, all rows are collapsed by default.)
        """

        model = self.sidebarModel

        frontier = model.rootNode.children[:]
        while frontier:
            node = frontier.pop()

            if not node.mayHaveChildren():
                continue

            if node.wantForceExpand() or node.getCollapseHash() not in model.collapseCache:
                index = self.nodeToFilterIndex(node)
                self.expand(index)

            frontier.extend(node.children)

    def expandChildFolders(self, node: SidebarNode, collapse: bool = False):
        operation = self.collapse if collapse else self.expand
        frontier = node.children[:]
        while frontier:
            node = frontier.pop()
            if node.kind == SidebarItem.RefFolder:
                frontier.extend(node.children)
                index = self.nodeToFilterIndex(node)
                operation(index)

    def collapseChildFolders(self, node: SidebarNode):
        self.expandChildFolders(node, collapse=True)

    def onIndexExpanded(self, index: QModelIndex):
        node = self.filterIndexToNode(index)
        self.sidebarModel.cacheNodeCollapsedState(node, collapsed=False)

    def onIndexCollapsed(self, index: QModelIndex):
        node = self.filterIndexToNode(index)
        self.sidebarModel.cacheNodeCollapsedState(node, collapsed=True)

    def setCollapseStateLayer(self, transient: bool):
        model = self.sidebarModel
        currentLayer = len(model.collapseCacheLayers) - 1
        targetLayer = 1 if transient else 0

        # No-op if switching to same layer
        if targetLayer == currentLayer:
            return

        if currentLayer == 0:
            # Create layer 1 (transient)
            model.collapseCacheLayers.append(set())
        elif currentLayer == 1:
            # Collapse everything first so restoreExpandedItems will work,
            # then drop down to layer 0 (permanent)
            self.collapseAll()
            model.collapseCacheLayers.pop()
        else:
            raise NotImplementedError(f"unsupported layer {currentLayer}")

        assert targetLayer == len(model.collapseCacheLayers) - 1

        self.restoreExpandedItems()

    # -------------------------------------------------------------------------

    def copyToClipboard(self, text: str):
        QApplication.clipboard().setText(text)
        self.statusMessage.emit(clipboardStatusMessage(text))

    def makeUpstreamSubmenu(
            self,
            repo: Repo,
            lbName: str,
            ubName: str,
            ubMissing: bool
    ) -> list[ActionDef]:
        radioGroup = "UpstreamSelection"

        menu = []

        # Missing upstream
        if ubMissing:
            assert ubName
            menu.append(ActionDef(
                ubName + " " + _("(Missing)"),
                checkState=1,
                icon="git-upstream-missing",
                enabled=False,
                radioGroup=radioGroup))

        # Stop tracking/Not tracking
        menu.append(ActionDef(
            _("Stop tracking upstream branch") if ubName else _("Not tracking any upstream branch"),
            lambda: EditUpstreamBranch.invoke(self, lbName, ""),
            checkState=1 if not ubName else -1,
            radioGroup=radioGroup))

        # All remote branches
        for remoteBranches in repo.listall_remote_branches(value_style="shorthand").values():
            if not remoteBranches:
                continue
            menu.append(ActionDef.SEPARATOR)
            for rbShorthand in sorted(remoteBranches, key=naturalSort):
                menu.append(ActionDef(
                    escamp(rbShorthand),
                    lambda rb=rbShorthand: EditUpstreamBranch.invoke(self, lbName, rb),
                    checkState=1 if ubName == rbShorthand else -1,
                    radioGroup=radioGroup))

        # "No remotes" label
        if len(menu) <= 1:
            if not repo.remotes:
                explainer = _("No remotes.")
            else:
                explainer = _("No remote branches found. Try fetching the remotes.")
            menu.append(ActionDef.SEPARATOR)
            menu.append(ActionDef(explainer, enabled=False))

        return menu

    def pushRefspecMenu(self, refspecs):
        remotes = self.sidebarModel.repoModel.remotes
        menu = []

        if remotes:
            menu.append(TaskBook.action(self, PushRefspecs, _("&All Remotes"), taskArgs=("*", refspecs)))
            menu.append(ActionDef.SEPARATOR)
            for remote in remotes:
                menu.append(TaskBook.action(self, PushRefspecs, escamp(remote), taskArgs=(remote, refspecs)))
        else:
            menu.append(ActionDef(_("No Remotes"), enabled=False))

        return menu



# -----------------------------------------------------------------------------
# Copyright (C) 2025 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.localization import *
from gitfourchette.qt import *
from gitfourchette.toolbox import stockIcon


class QHintButton(QToolButton):
    def __init__(self, parent=None, toolTip="", iconKey="hint"):
        super().__init__(parent)
        self.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.setAutoRaise(True)
        self.setText(_("Help"))
        self.setIcon(stockIcon(iconKey))
        self.setToolTip(toolTip)
        self.setCursor(Qt.CursorShape.WhatsThisCursor)
        self.connectClicked()

    def connectClicked(self):
        self.clicked.connect(lambda _: QToolTip.showText(QCursor.pos(), self.toolTip(), self))



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette import colors
from gitfourchette.localization import *
from gitfourchette.qt import *
from gitfourchette.toolbox import stockIcon, tquo


class CommitHookButton(QToolButton):
    def __init__(self, parent):
        super().__init__(parent)

        self.willRunHooks = True
        self.hookNames = []  # must be filled in

        toggleAction = QAction(self)
        toggleAction.setText("...RUN HOOKS...")
        toggleAction.setCheckable(True)
        toggleAction.setChecked(self.willRunHooks)
        toggleAction.toggled.connect(lambda t: self.updateConfig(t))
        self.toggleAction = toggleAction

        self.setPopupMode(QToolButton.ToolButtonPopupMode.InstantPopup)
        self.addActions([toggleAction])
        self.updateConfig(self.willRunHooks)

    def setup(self, commitHooks: list[str]):
        if not commitHooks:
            self.setVisible(False)
            return
        self.hookNames = commitHooks
        self.updateConfig(self.willRunHooks)

        title = _n("Run {h} hook", "Run hooks {h}", len(self.hookNames), h=self.hooksString())
        self.toggleAction.setText(title)

    def hooksString(self):
        andJoin = " " + _("and") + " "
        return andJoin.join([tquo(hook) for hook in self.hookNames])

    def updateConfig(self, willRunHooks: bool):
        self.willRunHooks = willRunHooks

        if willRunHooks:
            tip = _n("Hook {h} will run.", "Hooks {h} will run.", n=len(self.hookNames), h=self.hooksString())
            icon = stockIcon("git-hook", f"gray={colors.olive.name()}")
        else:
            tip = _n("Bypassing {h} hook for this commit.", "Bypassing hooks {h} for this commit.", n=len(self.hookNames), h=self.hooksString())
            icon = stockIcon("git-hook-disabled")

        self.setIcon(icon)
        self.setToolTip(tip)

    def explicitNoVerify(self):
        return not self.willRunHooks



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations

import logging
import shutil
from pathlib import Path
from tarfile import TarFile

import gitfourchette.pycompat  # noqa: F401 - Path.is_file(follow_symlinks=...) for Python 3.12
from gitfourchette import settings
from gitfourchette.qt import *
from gitfourchette.toolbox import withUniqueSuffix

logger = logging.getLogger(__name__)


class Trash:
    DirectoryName = "trash"
    QDateTimeFormat = "yyyyMMdd-HHmmss"
    MaxUniqueSuffix = 99

    _instance: Trash | None = None

    trashDir: Path
    trashFiles: list[Path]

    class BackupSkipped(Exception):
        pass

    def __init__(self):
        if not APP_TESTMODE:
            cacheDir = Path(QStandardPaths.writableLocation(QStandardPaths.StandardLocation.CacheLocation))
        else:
            # CacheLocation is common for all tests, but we don't want parallel
            # tests to pollute each other's trashes. So, use the test-specific
            # temporary directory. Put the trash into a subdirectory that needs
            # to be created in order to simulate a fresh install where the cache
            # directory doesn't exist yet.
            cacheDir = Path(qTempDir(), "fake_cache_directory")

        self.trashDir = Path(cacheDir, Trash.DirectoryName)
        self.trashFiles = []
        self.refreshFiles()

    @classmethod
    def instance(cls) -> Trash:
        if not cls._instance:
            cls._instance = cls()
        return cls._instance

    @classmethod
    def enabled(cls) -> bool:
        return cls.maxFileCount() > 0

    @classmethod
    def maxFileCount(cls) -> int:
        return settings.prefs.maxTrashFiles

    @classmethod
    def fileSizeLimit(cls) -> int:
        return settings.prefs.maxTrashFileKB * 1024

    @classmethod
    def hasFileSizeLimit(cls) -> bool:
        return settings.prefs.maxTrashFileKB > 0

    @classmethod
    def ensureTrashEnabled(cls):
        if not cls.enabled():
            raise Trash.BackupSkipped("trash disabled")

    def exists(self) -> bool:
        return self.trashDir.is_dir()

    def pathIsTrashManaged(self, p: Path) -> bool:
        if not p.is_relative_to(self.trashDir):
            return False
        return p.is_file(follow_symlinks=False) or p.is_symlink()

    def refreshFiles(self):
        self.trashFiles.clear()
        if self.exists():
            self.trashFiles.extend(p for p in self.trashDir.iterdir() if self.pathIsTrashManaged(p))
            self.trashFiles.sort(reverse=True)

    def makeRoom(self, maxFiles: int):
        while len(self.trashFiles) > maxFiles:
            f = self.trashFiles.pop()
            if self.pathIsTrashManaged(f):
                logger.debug(f"Deleting trash file {f}")
                f.unlink()

    def newFile(self, workdir: str, ext: str = "", originalPath: str = "") -> Path:
        self.ensureTrashEnabled()

        maxFiles = max(0, self.maxFileCount() - 1)
        self.makeRoom(maxFiles)

        self.trashDir.mkdir(parents=True, exist_ok=True)

        now = QDateTime.currentDateTime().toString(Trash.QDateTimeFormat)
        wdID = Path(workdir).name
        base = Path(originalPath).name
        stem = f"{now}-{wdID}---{base}"

        # If a file exists at this path, tack a number to the end of the name.
        uniqueName = withUniqueSuffix(
            stem,
            ext=ext,
            reserved=lambda candidate: Path(self.trashDir, candidate).exists(follow_symlinks=False),
            stop=Trash.MaxUniqueSuffix,
            suffixFormat="({})")

        path = Path(self.trashDir, uniqueName)

        # Replace existing file if withUniqueSuffix ran out of suffixes
        path.unlink(missing_ok=True)

        self.trashFiles.insert(0, path)
        return path

    def backupFile(self, workdir: str, originalPath: str) -> Path:
        self.ensureTrashEnabled()

        fullPath = Path(workdir, originalPath)
        try:
            size = fullPath.lstat().st_size
        except OSError as ex:  # FileNotFoundError, NotADirectoryError
            raise Trash.BackupSkipped("file inaccessible") from ex

        if self.hasFileSizeLimit() and size > self.fileSizeLimit():
            raise Trash.BackupSkipped("file too big")

        # Copy new file
        trashPath = self.newFile(workdir, originalPath=originalPath)
        shutil.copyfile(fullPath, trashPath, follow_symlinks=False)
        return trashPath

    def backupPatch(self, workdir: str, text: str, originalPath: str = "") -> Path:
        trashFile = self.newFile(workdir, ext=".patch", originalPath=originalPath)
        trashFile.write_text(text, encoding="utf-8", newline="\n")
        return trashFile

    def backupTree(self, workdir: str, treePath: str) -> Path:
        self.ensureTrashEnabled()

        treeFullPath = Path(workdir, treePath)

        # Check if tree is viable first
        if self.hasFileSizeLimit():
            totalSize = 0

            def reraise(walkError: OSError):
                raise Trash.BackupSkipped("inaccessible file in tree") from walkError

            for root, _dirs, files in treeFullPath.walk(on_error=reraise):
                for name in files:
                    fileSize = Path(root, name).lstat().st_size
                    totalSize += fileSize
                    if totalSize > self.fileSizeLimit():
                        raise Trash.BackupSkipped("tree too big")

        # Create a tarball
        trashFile = self.newFile(workdir, ext=".tar", originalPath=treePath)

        with TarFile(trashFile, "w") as tarball:
            tarball.add(treeFullPath, arcname=treePath)

        return trashFile

    def size(self) -> tuple[int, int]:
        size = 0
        count = 0

        for f in self.trashFiles:
            if self.pathIsTrashManaged(f):
                size += f.lstat().st_size
                count += 1

        return size, count

    def count(self) -> int:
        return self.size()[1]

    def clear(self):
        self.makeRoom(0)



# Form implementation generated from reading ui file 'ignorepatterndialog.ui'
#
# Created by: PyQt6 UI code generator 6.10.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_IgnorePatternDialog(object):
    def setupUi(self, IgnorePatternDialog):
        IgnorePatternDialog.setObjectName("IgnorePatternDialog")
        IgnorePatternDialog.resize(514, 131)
        self.formLayout = QFormLayout(IgnorePatternDialog)
        self.formLayout.setFieldGrowthPolicy(QFormLayout.FieldGrowthPolicy.ExpandingFieldsGrow)
        self.formLayout.setObjectName("formLayout")
        self.label = QLabel(parent=IgnorePatternDialog)
        self.label.setObjectName("label")
        self.formLayout.setWidget(0, QFormLayout.ItemRole.LabelRole, self.label)
        self.label_2 = QLabel(parent=IgnorePatternDialog)
        self.label_2.setObjectName("label_2")
        self.formLayout.setWidget(1, QFormLayout.ItemRole.LabelRole, self.label_2)
        spacerItem = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.formLayout.setItem(2, QFormLayout.ItemRole.FieldRole, spacerItem)
        self.buttonBox = QDialogButtonBox(parent=IgnorePatternDialog)
        self.buttonBox.setOrientation(Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Cancel|QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.formLayout.setWidget(3, QFormLayout.ItemRole.SpanningRole, self.buttonBox)
        self.fileEdit = QComboBoxWithPreview(parent=IgnorePatternDialog)
        self.fileEdit.setObjectName("fileEdit")
        self.formLayout.setWidget(1, QFormLayout.ItemRole.FieldRole, self.fileEdit)
        self.patternEdit = QComboBoxWithPreview(parent=IgnorePatternDialog)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.patternEdit.sizePolicy().hasHeightForWidth())
        self.patternEdit.setSizePolicy(sizePolicy)
        self.patternEdit.setEditable(True)
        self.patternEdit.setObjectName("patternEdit")
        self.formLayout.setWidget(0, QFormLayout.ItemRole.FieldRole, self.patternEdit)
        self.label.setBuddy(self.patternEdit)
        self.label_2.setBuddy(self.fileEdit)

        self.retranslateUi(IgnorePatternDialog)
        self.buttonBox.accepted.connect(IgnorePatternDialog.accept)
        self.buttonBox.rejected.connect(IgnorePatternDialog.reject)
        IgnorePatternDialog.setTabOrder(self.patternEdit, self.fileEdit)

    def retranslateUi(self, IgnorePatternDialog):
        IgnorePatternDialog.setWindowTitle(_p("IgnorePatternDialog", "Ignore file name pattern"))
        self.label.setText(_p("IgnorePatternDialog", "&Pattern:"))
        self.label_2.setText(_p("IgnorePatternDialog", "&Add to:"))
from gitfourchette.toolbox.qcomboboxwithpreview import QComboBoxWithPreview



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

"""
Library of widgets and utilities that aren't specifically tied to GitFourchette's core functionality.
"""

from .actiondef import ActionDef
from .autohidemenubar import AutoHideMenuBar
from .benchmark import Benchmark, benchmark
from .calledfromqthread import calledFromQThread
from .fittedtext import FittedText
from .excutils import shortenTracebackPath, excStrings
from .fontpicker import FontPicker
from .gitutils import (
    shortHash, nameValidationMessage,
    AuthorDisplayStyle, abbreviatePerson,
    PatchPurpose,
    remoteUrlProtocol,
    splitRemoteUrl,
    guessRemoteUrlFromText,
    formatTimeOffset,
    signatureEnvironmentVariables,
    signatureQDateTime,
    signatureDateFormat,
)
from .memoryindicator import MemoryIndicator
from .messageboxes import (
    MessageBoxIconName, excMessageBox, asyncMessageBox,
    showWarning, showInformation, askConfirmation,
    addULToMessageBox,
    NonCriticalOperation)
from .iconbank import stockIcon, stockIconImgTag
from .pathutils import PathDisplayStyle, abbreviatePath, compactPath, disambiguateTabTitlesByPath
from .persistentfiledialog import PersistentFileDialog
from .qbusyspinner import QBusySpinner
from .qcomboboxwithpreview import QComboBoxWithPreview
from .qelidedlabel import QElidedLabel
from .qfaintseparator import QFaintSeparator
from .qfilepickercheckbox import QFilePickerCheckBox
from .qhintbutton import QHintButton
from .qprocessconnection import QProcessConnection
from .qsignalconnectcontext import QSignalConnectContext
from .qsignalblockercontext import QSignalBlockerContext
from .qstatusbar2 import QStatusBar2
from .qtabwidget2 import QTabWidget2, QTabBar2
from .qtutils import (
    enforceComboBoxMaxVisibleItems,
    reevaluateStyleSheet,
    toggleQssProperty,
    isImageFormatSupported,
    onAppThread,
    setFontFeature,
    adjustedWidgetFontSize,
    tweakWidgetFont,
    formatWidgetText,
    formatWidgetTooltip,
    itemViewVisibleRowRange,
    isDarkTheme,
    mutedTextColorHex,
    mutedToolTipColorHex,
    appendShortcutToToolTipText,
    appendShortcutToToolTip,
    openFolder, showInFolder,
    DisableWidgetUpdatesContext,
    QScrollBackupContext,
    makeInternalLink,
    MultiShortcut,
    makeMultiShortcut,
    makeWidgetShortcut,
    CallbackAccumulator,
    lerp,
    mixColors,
    relativeLuminance,
    DocumentLinks,
    findParentWidget,
    setTabOrder,
    packDialog,
    QModelIndex_default,
    QPoint_zero,
)
from .textutils import (
    toLengthVariants,
    escape, escamp, paragraphs, messageSummary, elide,
    toRoomyUL,
    toTightUL,
    linkify,
    tagify,
    clipboardStatusMessage,
    hquo, hquoe, bquo, bquoe, lquo, lquoe, tquo, tquoe,
    btag,
    stripHtml,
    stripAccelerators,
    withUniqueSuffix,
    englishTitleCase,
    naturalSort,
    qstringLength,
)
from .urltooltip import UrlToolTip
from .validatormultiplexer import ValidatorMultiplexer
from .validators import ReplaceSpacesWithDashes



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.forms.brandeddialog import convertToBrandedDialog
from gitfourchette.qt import *
from gitfourchette.toolbox import QHintButton, ValidatorMultiplexer, setTabOrder


class TextInputDialog(QDialog):
    textAccepted = Signal(str)

    def __init__(
            self,
            parent: QWidget,
            title: str,
            label: str,
            subtitle: str = "",
            multilineSubtitle: bool = False,
            hint: str = "",
            selectableLabel: bool = True,
    ) -> None:
        super().__init__(parent)

        self.setWindowTitle(title)
        self.validator: ValidatorMultiplexer | None = None

        self.lineEdit = QLineEdit(self)

        self.buttonBox = QDialogButtonBox(self)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)

        layout = QGridLayout(self)

        if label:
            promptLabel = QLabel(label, parent=self)
            promptLabel.setTextFormat(Qt.TextFormat.AutoText)
            promptLabel.setWordWrap(True)
            layout.addWidget(promptLabel, 0, 0)
            if selectableLabel:
                promptLabel.setTextInteractionFlags(Qt.TextInteractionFlag.TextSelectableByMouse)

        if hint:
            hintButton = QHintButton(self, hint)
            self.buttonBox.addButton(hintButton, QDialogButtonBox.ButtonRole.HelpRole)

        layout.addWidget(self.lineEdit, 1, 0)
        layout.addWidget(self.buttonBox, 3, 0, 1, -1)
        # Leave row 2 free for setExtraWidget
        layout.setRowStretch(1, 1)
        self.contentsLayout = layout

        convertToBrandedDialog(self, subtitleText=subtitle, multilineSubtitle=multilineSubtitle)

        self.lineEdit.setFocus()

        # This size isn't guaranteed. But it'll expand the dialog horizontally if the label is shorter.
        self.setMinimumWidth(512)
        self.setWindowModality(Qt.WindowModality.WindowModal)

        # Connect signals
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.accepted.connect(lambda: self.textAccepted.emit(self.lineEdit.text()))

    def setText(self, text: str):
        assert self.validator is None, "initial text value must be set before installing the validator"
        self.lineEdit.setText(text)
        self.lineEdit.selectAll()

    def setValidator(self, validate: ValidatorMultiplexer.CallbackFunc):
        assert self.validator is None, "validator is already set!"
        validator = ValidatorMultiplexer(self)
        validator.setGatedWidgets(self.okButton)
        validator.connectInput(self.lineEdit, validate)
        validator.run()
        self.validator = validator

    def setExtraWidget(self, widget: QWidget):
        self.contentsLayout.addWidget(widget, 2, 0)
        setTabOrder(self.lineEdit, widget, self.buttonBox)

    @property
    def okButton(self) -> QPushButton:
        return self.buttonBox.button(QDialogButtonBox.StandardButton.Ok)

    @property
    def cancelButton(self) -> QPushButton:
        return self.buttonBox.button(QDialogButtonBox.StandardButton.Cancel)

    def show(self):
        super().show()
        self.setMaximumHeight(self.height())



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import sys as _sys
import os as _os
from typing import TYPE_CHECKING  # noqa: F401


def _envBool(key: str) -> bool:
    return _os.environ.get(key, "") not in ["", "0"]


APP_VERSION = "1.11.0"
APP_SYSTEM_NAME = "gitfourchette"
APP_DISPLAY_NAME = "GitFourchette"
APP_URL_SCHEME = APP_SYSTEM_NAME
APP_IDENTIFIER = "org.gitfourchette.gitfourchette"

# -----------------------------------------------------------------------------
# BEGIN_FREEZE_CONSTS
####################################
# Do not commit these changes!
####################################

APP_FREEZE_COMMIT = "9c30540d330e114f65493f89ee26db8b5a6c854d"
APP_FREEZE_DATE = "2026-08-21 15:02"
APP_FREEZE_QT = "pyqt6"
# END_FREEZE_CONSTS
# -----------------------------------------------------------------------------

APP_TESTMODE = _envBool("APP_TESTMODE") or "pytest" in _sys.modules
"""
Unit testing mode (don't touch real user prefs, etc.).
Can be forced with environment variable APP_TESTMODE.
"""

APP_DEBUG = APP_TESTMODE or _envBool("APP_DEBUG")
"""
Enable expensive assertions and debugging features.
Can be forced with environment variable APP_DEBUG.
Implied by APP_TESTMODE.
"""

APP_NOTHREADS = APP_TESTMODE or _envBool("APP_NOTHREADS")
"""
Disable multithreading (run all tasks on UI thread).
Can be forced with environment variable APP_NOTHREADS.
Implied by APP_TESTMODE.
"""

APP_VERBOSEDEL = _envBool("APP_VERBOSEDEL")
"""
Verbose QObject Python destructors (__del__).
"""

if APP_TESTMODE:
    APP_SYSTEM_NAME += "testmode"
    APP_DISPLAY_NAME += "TestMode"



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

"""
PyQt6/PySide6/PyQt5 compatibility layer
"""

# GitFourchette's preferred Qt binding is PyQt6, but you can use another
# binding via the QT_API environment variable. Values recognized by QT_API:
#       pyqt6
#       pyside6
#       pyqt5
# NOTE: PyQt5 support is deprecated and will be removed soon.
# PySide2 is not supported at all.
#
# If you're running unit tests, use the PYTEST_QT_API environment variable instead.
# If you're packaging the app, you may prefer to force a binding via appconsts.py.

import json as _json
import logging as _logging
import os as _os
import sys as _sys
from collections import abc as _collections_abc
from contextlib import suppress as _suppress

from gitfourchette.appconsts import *

_logger = _logging.getLogger(__name__)

_qtBindingOrder = ["pyqt6", "pyside6", "pyqt5"]

QT5 = False
QT6 = False
PYSIDE6 = False
PYQT5 = False
PYQT6 = False
MACOS = False
WINDOWS = False

if APP_FREEZE_QT:  # in frozen apps (PyInstaller, AppImage, Flatpak), target a fixed API
    _qtBindingOrder = [APP_FREEZE_QT]
    _qtBindingBootPref = _qtBindingOrder[0]
else:
    _qtBindingBootPref = _os.environ.get("QT_API", "").lower()

# If QT_API isn't set, see if the app's prefs file specifies a preferred Qt binding
if not _qtBindingBootPref:
    if _sys.platform == "darwin":
        _prefsPath = _os.path.expanduser("~/Library/Preferences")
    else:
        _prefsPath = _os.environ.get("XDG_CONFIG_HOME", _os.path.expanduser("~/.config"))
    _prefsPath = _os.path.join(_prefsPath, APP_SYSTEM_NAME, "prefs.json")
    with _suppress(OSError, ValueError):
        with open(_prefsPath, encoding="utf-8") as _f:
            _jsonPrefs = _json.load(_f)
        _qtBindingBootPref = _jsonPrefs.get("forceQtApi", "").lower()

if _qtBindingBootPref:
    if _qtBindingBootPref not in _qtBindingOrder:
        # Don't touch default binding order if user passed in an unsupported binding name.
        # Pass _qtBindingBootPref on to application code so it can complain.
        _logger.warning(f"Unrecognized Qt binding name: '{_qtBindingBootPref}'")
    else:
        # Move preferred binding to front of list
        _qtBindingOrder.remove(_qtBindingBootPref)
        _qtBindingOrder.insert(0, _qtBindingBootPref)

_logger.debug(f"Qt binding order is: {_qtBindingOrder}")

QT_BINDING = ""
QT_BINDING_VERSION = ""

def _bail(message: str):
    _sys.stderr.write(message + "\n")
    if QT_BINDING:
        _app = QApplication([])
        QMessageBox.critical(None, APP_DISPLAY_NAME, message)
    _sys.exit(1)

for _tentative in _qtBindingOrder:
    assert _tentative.islower()

    with _suppress(ImportError):
        if _tentative == "pyside6" and not TYPE_CHECKING:
            from PySide6.QtCore import *
            from PySide6.QtWidgets import *
            from PySide6.QtGui import *
            from PySide6 import __version__ as QT_BINDING_VERSION
            QT_BINDING = "PySide6"
            QT6 = PYSIDE6 = True
        elif _tentative == "pyqt6":
            from PyQt6.QtCore import *
            from PyQt6.QtWidgets import *
            from PyQt6.QtGui import *  # type: ignore[no-redef]
            QT_BINDING_VERSION = PYQT_VERSION_STR
            QT_BINDING = "PyQt6"
            QT6 = PYQT6 = True
        elif _tentative == "pyqt5" and not TYPE_CHECKING:
            from PyQt5.QtCore import *
            from PyQt5.QtWidgets import *
            from PyQt5.QtGui import *
            QT_BINDING_VERSION = PYQT_VERSION_STR
            QT_BINDING = "PyQt5"
            QT5 = PYQT5 = True
        else:
            _logger.warning(f"Unsupported Qt binding {_tentative}")

    if QT_BINDING:
        break  # We've successfully imported a binding, stop looking at candidates
else:
    _bail("No Qt binding found. Please install PyQt6 or PySide6.")

# -----------------------------------------------------------------------------
# Tweak initial environment variables

# Keep Qt from faking bold face with some variable fonts (see issue #10).
# This looks nicer out of the box in Ubuntu 24.10 and Fedora 41 (KDE spin).
# This is supposedly fixed in Qt 6.7 (https://bugreports.qt.io/browse/QTBUG-112136)
# but I've seen it occur with Qt 6.8 still.
_os.environ["QT_NO_SYNTHESIZED_BOLD"] = "1"

# -----------------------------------------------------------------------------
# Set up platform constants

QT_BINDING_BOOTPREF = _qtBindingBootPref
KERNEL = QSysInfo.kernelType().lower()
MACOS = KERNEL == "darwin"
WINDOWS = KERNEL == "winnt"
FREEDESKTOP = not MACOS and not WINDOWS
FLATPAK = FREEDESKTOP and _os.path.exists("/.flatpak-info")
FLATPAK_ID = _os.environ.get("FLATPAK_ID", "")
PYINSTALLER_MEIPASS = getattr(_sys, "_MEIPASS", "") if getattr(_sys, "frozen", False) else ""
XDG_RUNTIME_DIR = _os.environ.get("XDG_RUNTIME_DIR", "")
GNOME = "GNOME" in _os.environ.get("XDG_CURRENT_DESKTOP", "").upper().split(":")  # e.g. "ubuntu:GNOME"
KDE = "KDE" in _os.environ.get("XDG_CURRENT_DESKTOP", "").upper().split(":")
WAYLAND = _os.environ.get("XDG_SESSION_TYPE", "").upper() == "WAYLAND"
OFFSCREEN = _os.environ.get("QT_QPA_PLATFORM", "").upper() == "OFFSCREEN"

# Capture environment variables that were set on boot.
# These values will not be forwarded explicitly to subprocess environments.
INITIAL_ENVIRONMENT = _os.environ.copy()

# -----------------------------------------------------------------------------
# Try to import optional modules

# Test mode stuff
HAS_QTEST = False
with _suppress(ImportError):
    if PYQT6:
        from PyQt6.QtTest import QAbstractItemModelTester, QTest, QSignalSpy
    elif PYQT5 and not TYPE_CHECKING:
        from PyQt5.QtTest import QAbstractItemModelTester, QTest, QSignalSpy
    elif PYSIDE6 and not TYPE_CHECKING:
        from PySide6.QtTest import QAbstractItemModelTester, QTest, QSignalSpy
    HAS_QTEST = True

# Try to import QtDBus on Linux
HAS_QTDBUS = False
if FREEDESKTOP:
    with _suppress(ImportError):
        if PYQT6:
            from PyQt6.QtDBus import *
        elif PYQT5 and not TYPE_CHECKING:
            from PyQt5.QtDBus import *
        elif PYSIDE6 and not TYPE_CHECKING:
            from PySide6.QtDBus import *
        else:
            raise ImportError("QtDBus")
        HAS_QTDBUS = True

try:
    if PYQT6:
        from PyQt6.QtSvg import QSvgRenderer
    elif PYSIDE6 and not TYPE_CHECKING:
        from PySide6.QtSvg import QSvgRenderer
    elif PYQT5 and not TYPE_CHECKING:
        from PyQt5.QtSvg import QSvgRenderer
    else:
        raise ImportError("QtSvg")
except ImportError:
    _bail(f"{QT_BINDING} was found, but {QT_BINDING}'s bindings for QtSvg are missing.\n\nYour Linux distribution probably provides them as a separate package. Please install it and try again.")

# -----------------------------------------------------------------------------
# Patch some holes and incompatibilities in Qt bindings

# Match PyQt signal/slot names with PySide6
if PYQT5 or PYQT6:
    Signal = pyqtSignal
    SignalInstance = pyqtBoundSignal
    Slot = pyqtSlot

if PYSIDE6 and not TYPE_CHECKING:
    def _QCommandLineParser_addOptions(self, options):
        for o in options:
            self.addOption(o)

    QCommandLineParser.addOptions = _QCommandLineParser_addOptions

    # Unlike PyQt5/6, PySide6's QtWidgets module doesn't export QWIDGETSIZE_MAX
    try:
        _dummy = QWIDGETSIZE_MAX
    except NameError:
        QWIDGETSIZE_MAX = 16777215

if QT5 and not TYPE_CHECKING:
    # Disable "What's this?" in Qt 5 dialog box title bars (Qt 6 sets this off by default.)
    QCoreApplication.setAttribute(Qt.ApplicationAttribute.AA_DisableWindowContextHelpButton)

    # QMouseEvent.pos() is deprecated in Qt 6, so we don't use it.
    # Fill in QMouseEvent.position() for Qt 5.
    def _QMouseEvent_position(self: QMouseEvent) -> QPointF:
        return QPointF(self.pos())

    QMouseEvent.position = _QMouseEvent_position

    # Some of these QFontDatabase functions became static in Qt 6, but they weren't in Qt 5.
    _QFontDatabase_families = QFontDatabase.families
    _QFontDatabase_isFixedPitch = QFontDatabase.isFixedPitch
    _QFontDatabase_isPrivateFamily = QFontDatabase.isPrivateFamily
    QFontDatabase.families = lambda *_args, **_kwargs: _QFontDatabase_families(QFontDatabase(), *_args, **_kwargs)
    QFontDatabase.isFixedPitch = lambda *_args, **_kwargs: _QFontDatabase_isFixedPitch(QFontDatabase(), *_args, **_kwargs)
    QFontDatabase.isPrivateFamily = lambda *_args, **_kwargs: _QFontDatabase_isPrivateFamily(QFontDatabase(), *_args, **_kwargs)

# Return Consolas as default fixed font instead of Courier New on Windows.
if WINDOWS:
    def _QFontDatabase_systemFont_override(type: QFontDatabase.SystemFont):
        font = _QFontDatabase_systemFont(type)
        if type == QFontDatabase.SystemFont.FixedFont:
            font.setFamily("Consolas")
        return font

    _QFontDatabase_systemFont = QFontDatabase.systemFont
    QFontDatabase.systemFont = _QFontDatabase_systemFont_override  # type: ignore[method-assign]

# Qt 6.7 replaces QCheckBox.stateChanged with checkStateChanged.
if not hasattr(QCheckBox, 'checkStateChanged'):
    # Note: this forwards an int, not a real CheckState, but the values are the same.
    QCheckBox.checkStateChanged = QCheckBox.stateChanged

# Qt 6.6 introduces QPalette.accent
if not hasattr(QPalette, 'accent'):
    QPalette.accent = lambda *_args, **_kwargs: QBrush(QColor(0x1da0d3))  # type: ignore[method-assign]

# Pythonic iterator for QTextFragments in a QTextBlock. Use this instead of QTextBlock.__iter__,
# which in PySide6 is an inconvenient QTextBlock::iterator, and in PyQt6 isn't implemented at all.
def _QTextBlock_fragments(block: QTextBlock) -> _collections_abc.Iterator[QTextFragment]:
    iterator = block.begin()  # QTextBlock::iterator
    while not iterator.atEnd():
        fragment = iterator.fragment()
        if fragment.isValid():
            yield iterator.fragment()
        iterator += 1

QTextBlock.fragments = _QTextBlock_fragments  # type: ignore[attr-defined]

# Custom "selected, no focus" icon mode.
QIcon.Mode.SelectedInactive = QIcon.Mode(4)  # type: ignore[attr-defined]


# -----------------------------------------------------------------------------
# Verbose QObject destructors

if APP_VERBOSEDEL:
    def _QObject_dtor(obj: QObject):
        cls = obj.__class__
        if cls.__module__.startswith(QT_BINDING + "."):  # Skip standard Qt objects
            return

        message = f"QObject.__del__ {cls.__name__}"
        app = QApplication.instance()
        if app and app.thread() is not QThread.currentThread():
            message += " !!!DANGER!!! QObject destroyed outside of UI thread!"

        _logger.info(message)

    assert not hasattr(QObject, "__del__")
    QObject.__del__ = _QObject_dtor  # type: ignore[attr-defined]
    _logger.info("Verbose QObject destructors enabled")


# -----------------------------------------------------------------------------
# Utility functions

def qAppName() -> str:
    """ User-facing application name. Shorthand for QApplication.applicationDisplayName(). """
    return QApplication.applicationDisplayName()


def qTempDir() -> str:
    """ Path to temporary directory for this session. """
    return QApplication.instance().tempDir.path()  # type: ignore[attr-defined]



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations

import warnings
from typing import Any, ClassVar

from gitfourchette import tasks
from gitfourchette.localization import *
from gitfourchette.qt import *
from gitfourchette.tasks import RepoTask
from gitfourchette.toolbox import MultiShortcut, makeMultiShortcut, ActionDef, englishTitleCase


class TaskBook:
    """ Registry of metadata about task commands """

    names       : ClassVar[dict[type[RepoTask], str]] = {}
    toolbarNames: ClassVar[dict[type[RepoTask], str]] = {}
    tips        : ClassVar[dict[type[RepoTask], str]] = {}
    shortcuts   : ClassVar[dict[type[RepoTask], MultiShortcut]] = {}
    icons       : ClassVar[dict[type[RepoTask], str]] = {}
    noEllipsis  : ClassVar[set[type[RepoTask]]] = set()

    @classmethod
    def retranslate(cls):
        cls.names = {
            tasks.AbortMerge: _("Abort merge"),
            tasks.AbsorbSubmodule: _("Absorb submodule"),
            tasks.AutoFetchRemotes: _("Auto-fetch remotes"),
            tasks.AcceptMergeConflictResolution: _("Accept merge conflict resolution"),
            tasks.AmendCommit: _("Amend last commit"),
            tasks.ApplyPatch: _("Apply selected text"),
            tasks.ApplyPatchData: _("Apply patch data"),
            tasks.ApplyPatchFile: _("Apply patch file"),
            tasks.ApplyPatchFileReverse: _("Revert patch file"),
            tasks.ApplyStash: _("Apply stash"),
            tasks.CheckoutCommit: _("Check out commit"),
            tasks.CherrypickCommit: _("Cherry-pick"),
            tasks.DeleteBranch: _("Delete local branch"),
            tasks.DeleteBranchFolder: _("Delete local branch folder"),
            tasks.DeleteRemote: _("Remove remote"),
            tasks.DeleteRemoteBranch: _("Delete branch on remote"),
            tasks.DeleteTag: _("Delete tag"),
            tasks.DiscardFiles: _("Discard changes"),
            tasks.DiscardModeChanges: _("Discard mode changes"),
            tasks.DownloadLfsObjects: _("Download LFS objects"),
            tasks.DropStash: _("Delete stash"),
            tasks.EditRemote: _("Edit remote"),
            tasks.EditUpstreamBranch: _("Edit upstream branch"),
            tasks.ExportCommitAsPatch: _("Export commit as patch file"),
            tasks.ExportPatchCollection: _("Export patch file"),
            tasks.ExportStashAsPatch: _("Export stash as patch file"),
            tasks.ExportWorkdirAsPatch: _("Export changes as patch file"),
            tasks.FastForwardBranch: _("Fast-forward branch"),
            tasks.FetchRemotes: _("Fetch remote branches"),
            tasks.FetchRemoteBranch: _("Fetch remote branch"),
            tasks.GetCommitInfo: _("Get commit information"),
            tasks.HardSolveConflicts: _("Accept/reject incoming changes"),
            tasks.Jump: _("Navigate in repo"),
            tasks.JumpBack: _("Navigate back"),
            tasks.JumpForward: _("Navigate forward"),
            tasks.JumpToHEAD: _("Go to HEAD commit"),
            tasks.JumpToUncommittedChanges: _("Go to Working Directory"),
            tasks.LoadPatchInNewWindow: _("Open diff in new window"),
            tasks.MergeBranch: _("Merge branch"),
            tasks.NewBranchFromCommit: _("New local branch"),
            tasks.NewBranchFromHead: _("New local branch"),
            tasks.NewBranchFromRef: _("New local branch"),
            tasks.NewCommit: _("New commit"),
            tasks.NewIgnorePattern: _("Ignore a new file name pattern"),
            tasks.NewRemote: _("Add remote"),
            tasks.NewStash: _("Stash changes"),
            tasks.NewTag: _("New tag"),
            tasks.OpenBlame: _("Blame file"),
            tasks.OpenInDiffTool: _("Open in external diff tool"),
            tasks.OpenMergeTool: _("Open in merge tool"),
            tasks.OpenRevisionInEditor: _("Open file revision"),
            tasks.QueryCommitsTouchingPath: _("Find commits touching path"),
            tasks.PullBranch: _("Pull remote branch"),
            tasks.PushBranch: _("Push branch"),
            tasks.PushRefspecs: _("Push refspecs"),
            tasks.RecallCommit: _("Recall lost commit"),
            tasks.RefreshRepo: _("Refresh repo"),
            tasks.RegisterSubmodule: _("Register submodule"),
            tasks.RemoveSubmodule: _("Remove submodule"),
            tasks.RenameBranch: _("Rename local branch"),
            tasks.RenameBranchFolder: _("Rename local branch folder"),
            tasks.RenameRemoteBranch: _("Rename branch on remote"),
            tasks.ResetHead: _("Reset HEAD"),
            tasks.RestoreRevisionToWorkdir: _("Restore file revision"),
            tasks.RevertCommit: _("Revert commit"),
            tasks.SetUpGitIdentity: _("Git identity"),
            tasks.EditRepoSettings: _("Repository settings"),
            tasks.SaveRevisionAs: _("Save file revision as"),
            tasks.StageFiles: _("Stage files"),
            tasks.SwitchBranch: _("Switch to branch"),
            tasks.UpdateSubmodule: _("Update submodule"),
            tasks.UpdateSubmodulesRecursive: _("Update submodules recursively"),
            tasks.UnstageFiles: _("Unstage files"),
            tasks.UnstageModeChanges: _("Unstage mode changes"),
            tasks.VerifyGpgSignature: _("Verify signature"),
            tasks.VerifyGpgQueue: _("Verifying signatures"),
        }

        cls.toolbarNames = {
            tasks.AmendCommit: _p("toolbar", "Amend"),
            tasks.FetchRemotes: _p("toolbar", "Fetch"),
            tasks.JumpBack: _p("toolbar", "Back"),
            tasks.JumpForward: _p("toolbar", "Forward"),
            tasks.JumpToHEAD: _p("toolbar", "HEAD"),
            tasks.JumpToUncommittedChanges: _p("toolbar", "Workdir"),
            tasks.NewBranchFromHead: _p("toolbar", "Branch"),
            tasks.NewStash: _p("toolbar", "Stash"),
            tasks.PullBranch: _p("toolbar", "Pull"),
            tasks.PushBranch: _p("toolbar", "Push"),
        }

        cls.tips = {
            tasks.AmendCommit: _("Amend the last commit on the current branch with the staged changes in the working directory"),
            tasks.ApplyPatchFile: _("Apply a patch file to the working directory"),
            tasks.ApplyPatchFileReverse: _("Apply a patch file (in reverse) to the working directory"),
            tasks.ApplyStash: _("Restore backed up changes to the working directory"),
            tasks.CherrypickCommit: _("Bring the changes introduced by this commit to the current branch"),
            tasks.DeleteBranch: _("Delete this branch locally"),
            tasks.EditUpstreamBranch: _("Choose the remote branch to be tracked by this local branch"),
            tasks.ExportStashAsPatch: _("Create a patch file from this stash"),
            tasks.FastForwardBranch: _("Advance this local branch to the tip of the remote-tracking branch"),
            tasks.FetchRemotes: _("Get the latest commits on all remote branches"),
            tasks.FetchRemoteBranch: _("Get the latest commits from the remote server"),
            tasks.NewBranchFromCommit: _("Start a new branch from this commit"),
            tasks.NewBranchFromHead: _("Start a new branch from the current HEAD"),
            tasks.NewBranchFromRef: _("Start a new branch from the tip of this branch"),
            tasks.NewCommit: _("Create a commit of the staged changes in the working directory"),
            tasks.NewRemote: _("Add a remote server to this repo"),
            tasks.NewStash: _("Back up uncommitted changes and clean up the working directory"),
            tasks.NewTag: _("Tag this commit with a name"),
            tasks.PullBranch: _("Fetch the latest commits from the remote, then integrate them into your local branch"),
            tasks.PushBranch: _("Upload your commits on the current branch to the remote server"),
            tasks.RemoveSubmodule: _("Remove this submodule from .gitmodules and delete its working copy from this repo"),
            tasks.RenameBranch: _("Rename this branch locally"),
            tasks.ResetHead: _("Make HEAD point to another commit"),
            tasks.RevertCommit: _("Revert the changes introduced by this commit"),
            tasks.SetUpGitIdentity: _("Set up the identity under which you create commits"),
            tasks.EditRepoSettings: _("Set up the identity under which you create commits"),
            tasks.SwitchBranch: _("Switch to this branch and update the working directory to match it"),
        }

    @classmethod
    def initialize(cls):
        cls.names = {}
        cls.toolbarNames = {}
        cls.tips = {}

        cls.shortcuts = {
            tasks.AmendCommit: makeMultiShortcut(QKeySequence.StandardKey.SaveAs, "Ctrl+Shift+S"),
            tasks.ApplyPatchFile: makeMultiShortcut("Ctrl+I"),
            tasks.FetchRemotes: makeMultiShortcut("Ctrl+Shift+R"),
            tasks.JumpBack: makeMultiShortcut("Ctrl+Left" if MACOS else "Alt+Left"),
            tasks.JumpForward: makeMultiShortcut("Ctrl+Right" if MACOS else "Alt+Right"),

            # macOS: ⌘H is reserved by the window manager, use ^H instead.
            # macOS: For consistency with ^H, use ^G (not ⌘G) to jump to the workdir.
            tasks.JumpToHEAD: makeMultiShortcut("Meta+H" if MACOS else "Ctrl+H"),
            tasks.JumpToUncommittedChanges: makeMultiShortcut("Meta+G" if MACOS else "Ctrl+G"),

            tasks.NewBranchFromHead: makeMultiShortcut("Ctrl+B"),
            tasks.NewCommit: makeMultiShortcut(QKeySequence.StandardKey.Save),
            tasks.NewStash: makeMultiShortcut("Ctrl+Alt+S"),
            tasks.OpenBlame: makeMultiShortcut("Ctrl+L"),
            tasks.PullBranch: makeMultiShortcut("Ctrl+Shift+P"),
            tasks.PushBranch: makeMultiShortcut("Ctrl+P"),
        }

        cls.icons = {
            tasks.AmendCommit: "git-commit-amend",
            tasks.CheckoutCommit: "git-checkout",
            tasks.CherrypickCommit: "git-cherrypick",
            tasks.DeleteBranch: "vcs-branch-delete",
            tasks.DeleteRemote: "SP_TrashIcon",
            tasks.DeleteRemoteBranch: "SP_TrashIcon",
            tasks.DeleteTag: "SP_TrashIcon",
            tasks.DropStash: "SP_TrashIcon",
            tasks.EditRemote: "document-edit",
            tasks.EditRepoSettings: "configure",
            tasks.FetchRemoteBranch: "git-fetch",
            tasks.FetchRemotes: "git-fetch",
            tasks.JumpBack: "back",
            tasks.JumpForward: "forward",
            tasks.JumpToHEAD: "git-head",
            tasks.JumpToUncommittedChanges: "git-workdir",
            tasks.MergeBranch: "git-merge",
            tasks.NewBranchFromCommit: "git-branch",
            tasks.NewBranchFromHead: "git-branch",
            tasks.NewBranchFromRef: "git-branch",
            tasks.NewCommit: "git-commit",
            tasks.NewRemote: "git-remote",
            tasks.NewStash: "git-stash-black",
            tasks.NewTag: "git-tag",
            tasks.OpenBlame: "git-blame",
            tasks.PullBranch: "git-pull",
            tasks.PushBranch: "git-push",
            tasks.SetUpGitIdentity: "user-identity",
            tasks.SwitchBranch: "git-checkout",
        }

        cls.noEllipsis = {
            tasks.FastForwardBranch,
            tasks.FetchRemotes,
            tasks.FetchRemoteBranch,
            tasks.JumpBack,
            tasks.JumpForward,
            tasks.JumpToHEAD,
            tasks.JumpToUncommittedChanges,
            tasks.UnstageModeChanges,
            tasks.UpdateSubmodule,
            tasks.UpdateSubmodulesRecursive,
        }

        cls.retranslate()

    @classmethod
    def autoActionName(cls, t: type[RepoTask]):
        assert cls.names
        try:
            name = cls.names[t]
        except KeyError:
            name = t.__name__
            cls.names[t] = name
            warnings.warn(f"Missing name for task '{name}'")

        name = englishTitleCase(name)

        if t not in cls.noEllipsis:
            name += "…"

        return name

    @classmethod
    def action(
            cls,
            invoker: QObject,
            taskClass: type[RepoTask],
            name="",
            accel="",
            taskArgs: Any = None,
            **kwargs
    ) -> ActionDef:
        if not name:
            name = cls.autoActionName(taskClass)

        if accel:
            name = cls.autoActionName(taskClass)
            i = name.lower().find(accel.lower())
            if i >= 0:
                name = name[:i] + "&" + name[i:]

        if taskArgs is None:
            taskArgs = ()
        elif not isinstance(taskArgs, tuple | list):
            taskArgs = (taskArgs,)

        icon = cls.icons.get(taskClass, "")
        shortcuts = cls.shortcuts.get(taskClass, [])
        tip = cls.tips.get(taskClass, "")

        def callback():
            taskClass.invoke(invoker, *taskArgs)

        actionDef = ActionDef(name, callback=callback, icon=icon, shortcuts=shortcuts, tip=tip)

        if kwargs:
            actionDef = actionDef.replace(**kwargs)

        return actionDef

    @classmethod
    def toolbarAction(cls, invoker: QObject, taskClass: type[RepoTask]):
        name = cls.toolbarNames.get(taskClass, "")
        tip = cls.autoActionName(taskClass)
        return cls.action(invoker, taskClass, name).replace(tip=tip)



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from math import ceil
from typing import ClassVar

from gitfourchette.qt import *

class FittedText:
    enable = True

    stretchPresets: ClassVar = [
        QFont.Stretch.SemiCondensed,  # 87
        QFont.Stretch.Condensed,  # 75
        QFont.Stretch(70),
        QFont.Stretch(66),
        QFont.Stretch.ExtraCondensed,  # 62
        QFont.Stretch(56),
        QFont.Stretch.UltraCondensed,  # 50
    ]

    if MACOS:
        # WEIRD! SemiCondensed is actually wider than Unstretched
        # with Mac system font (Qt 6.8.0, macOS 15)
        stretchPresets.remove(QFont.Stretch.SemiCondensed)

    defaultStretchLimit = stretchPresets[0]

    @classmethod
    def fit(
            cls,
            wideFont: QFont,
            maxWidth: int,
            text: str,
            mode=Qt.TextElideMode.ElideRight,
            limit=defaultStretchLimit,
            bypassSetting=False,
    ) -> tuple[str, QFont, int]:
        metrics = QFontMetricsF(wideFont)
        width = ceil(metrics.horizontalAdvance(text))

        if width < 1:
            return text, wideFont, 0

        if not cls.enable and not bypassSetting:
            font = wideFont
        else:
            # Figure out upper bound for the stretch factor
            baselineStretch = int(100 * maxWidth / width)

            if baselineStretch >= 100:
                # No condensing needed - early out
                return text, wideFont, width

            baselineStretch = max(baselineStretch, limit)

            font = QFont(wideFont)
            for stretch in cls.stretchPresets:
                # Skip stretch factors that are too wide
                if stretch > baselineStretch:
                    continue

                # Don't stretch beyond user preference
                if stretch < limit:
                    break

                font.setStretch(stretch)
                metrics = QFontMetricsF(font)

                width = ceil(metrics.horizontalAdvance(text))

                if width < maxWidth:
                    return text, font, width

        # Still no room at most condensed preset, must elide the text
        text = metrics.elidedText(text, mode, maxWidth)
        width = ceil(metrics.horizontalAdvance(text))
        return text, font, width

    @classmethod
    def draw(
            cls,
            painter: QPainter,
            rect: QRect,
            flags: Qt.AlignmentFlag,
            text: str,
            mode=Qt.TextElideMode.ElideRight,
            minStretch=defaultStretchLimit,
            bypassSetting=False,
    ) -> tuple[str, QFont, int]:
        wideFont = painter.font()
        text, font, width = cls.fit(wideFont, rect.width(), text, mode, minStretch, bypassSetting)
        if font is wideFont:
            painter.drawText(rect, flags, text)
        else:
            painter.setFont(font)
            painter.drawText(rect, flags, text)
            painter.setFont(wideFont)
        return text, font, width



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations

import difflib
import re
from bisect import bisect_left
from collections.abc import Iterator
from dataclasses import dataclass

from gitfourchette import colors
from gitfourchette import settings
from gitfourchette.gitdriver.parsers import iterateLines
from gitfourchette.localization import *
from gitfourchette.qt import *
from gitfourchette.syntax import LexJob, ColorScheme
from gitfourchette.toolbox import *


_indexLinePattern = re.compile(r"index ([\da-f]+)\.\.([\da-f]+)")
_hunkHeaderPattern = re.compile(r"@@ -(\d+)(?:,(\d+))? \+(\d+)(?:,(\d+))? @@")


def _parseHunkHeader(text: str) -> tuple[int, int, int, int, str]:
    match = _hunkHeaderPattern.match(text)
    oldStartStr, oldLinesStr, newStartStr, newLinesStr = match.groups()
    oldStart = int(oldStartStr)
    newStart = int(newStartStr)
    oldLines = int(oldLinesStr) if oldLinesStr else 1
    newLines = int(newLinesStr) if newLinesStr else 1
    comment = text[match.end():]
    return oldStart, oldLines, newStart, newLines, comment


@dataclass
class DiffLinePos:
    hunkID: int

    hunkLineNum: int
    """
    0-indexed line number within the hunk.
    -1 means the hunk header line itself ("@@ ... @@").
    """

    def isHunkHeaderLine(self) -> bool:
        return self.hunkLineNum == -1


@dataclass
class LineData:
    text: str
    "Line text for visual representation."

    hunkPos: DiffLinePos
    "Which hunk this line pertains to, and its position in the hunk."

    origin: str = ""
    oldLineNo: int = -1
    newLineNo: int = -1

    cursorStart: int = -1
    "Cursor position at start of line in QDocument."

    cursorEnd: int = -1
    "Cursor position at end of line in QDocument."

    clumpID: int = -1
    "Which clump this line pertains to. 'Clumps' are groups of adjacent +/- lines."

    doppelganger: int = -1
    "Index of the doppelganger LineData in a perfectly even clump."

    trailerLength: int = 0
    "Stop highlighting the syntax past this column in the line."

    hiddenSuffix: str = ""
    """
    Suffix that follows `text` in a patch file but should be hidden in the UI,
    e.g. newline + backslash + ' No newline at end of file'.
    """

    @classmethod
    def getHunkExtents(cls, array: list[LineData], hunkID: int) -> tuple[int, int]:
        """
        Find indices of first and last LineData objects given the current hunk.
        """

        first = bisect_left(array, hunkID, 0, key=LineData._hunkIDKey)
        last = bisect_left(array, hunkID + 1, first, key=LineData._hunkIDKey) - 1
        return first, last

    def _hunkIDKey(self) -> int:
        return self.hunkPos.hunkID

    def parseHunkHeader(self) -> tuple[int, int, int, int, str]:
        assert self.hunkPos.hunkLineNum == -1
        return _parseHunkHeader(self.text)

    @property
    def originDelta(self) -> int:
        if self.origin == "+":
            return 1
        if self.origin == "-":
            return -1
        return 0

    @property
    def reverseOrigin(self) -> str:
        if self.origin == "+":
            return "-"
        if self.origin == "-":
            return "+"
        return self.origin


class DiffTextFormats:
    addBF = QTextBlockFormat()
    delBF = QTextBlockFormat()
    doppelgangerAddCF = QTextCharFormat()
    doppelgangerDelCF = QTextCharFormat()
    hunkBF = QTextBlockFormat()
    hunkCF = QTextCharFormat()
    warningCF = QTextCharFormat()

    @classmethod
    def refresh(cls, scheme: ColorScheme, colorblind: bool):
        bgColor = scheme.backgroundColor

        # Base del/add backgrounds.
        # Instead of using an alpha value, blend with the background color.
        # This way, we avoid overlapping alpha artifacts at the edges of QCharFormat runs.
        if colorblind:
            delColor1 = mixColors(bgColor, colors.orange, .35)
            addColor1 = mixColors(bgColor, colors.teal, .35)
        else:
            delColor1 = mixColors(bgColor, QColor(0xff5555), .35)
            addColor1 = mixColors(bgColor, QColor(0x55ff55), .35)

        # Starker contrast for per-character highlights
        if colorblind:
            delColor2 = mixColors(delColor1, colors.orange, .6)
            addColor2 = mixColors(addColor1, colors.teal, .6)
        elif scheme.isDark():
            delColor2 = mixColors(delColor1, QColor(0x993333), .6)
            addColor2 = mixColors(addColor1, QColor(0x339933), .6)
        else:
            delColor2 = mixColors(delColor1, QColor(0x993333), .25)
            addColor2 = mixColors(addColor1, QColor(0x339933), .25)

        if scheme.isDark():
            warningColor = colors.red
            hunkColor = colors.blue.lighter(125)
        else:
            warningColor = colors.red.darker(125)
            hunkColor = QColor(0x0050f0)

        cls.addBF.setBackground(addColor1)
        cls.delBF.setBackground(delColor1)

        cls.doppelgangerAddCF.setBackground(addColor2)
        cls.doppelgangerDelCF.setBackground(delColor2)

        cls.hunkCF.setFontItalic(True)
        cls.hunkCF.setForeground(hunkColor)

        cls.warningCF.setFontUnderline(True)
        cls.warningCF.setFontWeight(QFont.Weight.Bold)
        cls.warningCF.setBackground(scheme.backgroundColor)
        cls.warningCF.setForeground(warningColor)


@dataclass
class DiffDocument:
    document: QTextDocument
    lineData: list[LineData]
    pluses: int
    minuses: int
    maxLine: int
    oldHash: str
    newHash: str

    # Syntax highlighting
    oldLexJob: LexJob | None = None
    newLexJob: LexJob | None = None

    class VeryLongLinesError(ValueError):
        pass

    class NoChangeError(ValueError):
        pass

    class BinaryError(ValueError):
        pass

    @staticmethod
    def fromPatch(patch: str, maxLineLength=0) -> DiffDocument:
        lineData: list[LineData] = []

        clumpID = 0
        numLinesInClump = 0
        perfectClumpTally = 0
        pluses = 0
        minuses = 0

        hunkID = -1
        oldLine = -1
        newLine = -1
        hunkLineNum = -1
        isBinary = False
        oldHash = ""
        newHash = ""

        for pos, endPos in iterateLines(patch):
            if maxLineLength and endPos - pos > maxLineLength:
                raise DiffDocument.VeryLongLinesError()

            firstChar = patch[pos]

            # Keep looking for first hunk
            if firstChar != "@" and hunkID < 0:
                if patch.startswith(("Binary files", "GIT binary patch"), pos, endPos):
                    isBinary = True
                elif patch.startswith("index ", pos, endPos):
                    # The 'index' line can be used to complete an existing delta with actual blob hashes
                    indexLineMatch = _indexLinePattern.match(patch, pos, endPos)
                    oldHash, newHash = indexLineMatch.groups()
                continue

            # Start new hunk
            if firstChar == "@":
                rawLine = patch[pos:endPos]
                oldLine, _dummy1, newLine, _dummy2, _dummy3 = _parseHunkHeader(rawLine)

                hunkID += 1
                hunkLineNum = -1
                hunkHeaderLD = LineData(text=rawLine, hunkPos=DiffLinePos(hunkID, -1),
                                        oldLineNo=oldLine, newLineNo=newLine)
                lineData.append(hunkHeaderLD)
                continue

            origin = firstChar

            # "No newline at end of file" (message might be localized)
            if origin == '\\':
                # Fix up the last LineData and don't create a new one.
                ld = lineData[-1]
                ld.text = ld.text.removesuffix("\n")
                ld.hiddenSuffix = "\n" + patch[pos:endPos]
                continue

            hunkLineNum += 1

            # Any lines that aren't +/- break up the current clump
            if origin not in "+-" and numLinesInClump != 0:
                # Process perfect clump (sum of + and - origins is 0)
                if numLinesInClump > 0 and perfectClumpTally == 0:
                    assert (numLinesInClump % 2) == 0, "line count should be even in perfect clumps"
                    clumpStart = len(lineData) - numLinesInClump
                    halfClump = numLinesInClump // 2
                    for doppel1 in range(clumpStart, clumpStart + halfClump):
                        doppel2 = doppel1 + halfClump
                        lineData[doppel1].doppelganger = doppel2
                        lineData[doppel2].doppelganger = doppel1

                # Start new clump
                clumpID += 1
                numLinesInClump = 0
                perfectClumpTally = 0

            ld = LineData(text=patch[pos+1:endPos],
                          hunkPos=DiffLinePos(hunkID, hunkLineNum),
                          origin=origin,
                          oldLineNo=-1 if origin == "+" else oldLine,
                          newLineNo=-1 if origin == "-" else newLine)

            if origin == '+':
                assert ld.newLineNo == newLine
                assert ld.oldLineNo == -1
                newLine += 1
                ld.clumpID = clumpID
                numLinesInClump += 1
                perfectClumpTally += 1
                pluses += 1
            elif origin == '-':
                assert ld.newLineNo == -1
                assert ld.oldLineNo == oldLine
                oldLine += 1
                ld.clumpID = clumpID
                numLinesInClump += 1
                perfectClumpTally -= 1
                minuses += 1
            else:
                assert origin == " ", f"unknown origin: {origin.encode('unicode_escape')!r}"
                assert ld.newLineNo == newLine
                assert ld.oldLineNo == oldLine
                newLine += 1
                oldLine += 1

            lineData.append(ld)

        if not lineData:
            if isBinary:
                raise DiffDocument.BinaryError()
            raise DiffDocument.NoChangeError()

        # Recreating a QTextDocument is faster than clearing any existing one.
        textDocument = QTextDocument()
        textDocument.setObjectName("DiffDocument")
        textDocument.setDocumentLayout(QPlainTextDocumentLayout(textDocument))

        diffDocument = DiffDocument(document=textDocument, lineData=lineData,
                                    pluses=pluses, minuses=minuses,
                                    maxLine=max(newLine, oldLine),
                                    oldHash=oldHash, newHash=newHash)

        # Begin batching text insertions for performance.
        # This prevents Qt from recomputing the document's layout after every line insertion.
        cursor = QTextCursor(textDocument)
        cursor.beginEditBlock()

        # Build up document from the lineData array.
        diffDocument.buildTextDocument(cursor)

        # Emphasize doppelganger differences.
        diffDocument.formatDoppelgangerDiffs(cursor)

        # Done batching text insertions.
        cursor.endEditBlock()

        return diffDocument

    @benchmark
    def buildTextDocument(self, cursor: QTextCursor):
        assert self.document.isEmpty()

        defaultBF = cursor.blockFormat()
        defaultCF = cursor.charFormat()
        showStrayCRs = settings.prefs.showStrayCRs
        isEmpty = True

        for ld in self.lineData:
            # Decide block format & character format
            origin = ld.origin
            if not origin:
                bf = DiffTextFormats.hunkBF
                cf = DiffTextFormats.hunkCF
            elif origin == '+':
                bf = DiffTextFormats.addBF
                cf = defaultCF
            elif origin == '-':
                bf = DiffTextFormats.delBF
                cf = defaultCF
            else:
                bf = defaultBF
                cf = defaultCF

            # Process line ending
            trailer = ""
            if ld.text.endswith('\r\n'):
                trimBack = -2
                if showStrayCRs:
                    trailer = "<CRLF>"
            elif ld.text.endswith('\n'):
                trimBack = -1
            elif ld.text.endswith('\r'):
                trimBack = -1
                if showStrayCRs:
                    trailer = "<CR>"
            else:
                trailer = _("<no newline at end of file>")
                trimBack = None  # yes, None. This will cancel slicing.

            if isEmpty:
                ld.cursorStart = 0
            else:
                cursor.insertBlock()
                ld.cursorStart = cursor.position()

            cursor.setBlockFormat(bf)
            cursor.setBlockCharFormat(cf)
            cursor.insertText(ld.text[:trimBack])

            if trailer:
                ld.trailerLength = len(trailer)
                cursor.setCharFormat(DiffTextFormats.warningCF)
                cursor.insertText(trailer)

            ld.cursorEnd = cursor.position()
            isEmpty = False

    @benchmark
    def formatDoppelgangerDiffs(self, cursor: QTextCursor):
        if self.pluses == 0 or self.minuses == 0:  # Don't bother if there can't be any doppelgangers
            return

        doppelgangerBlocksQueue = []

        for lineNumber, line in enumerate(self.lineData):
            if line.doppelganger < 0:  # Skip lines without doppelgangers
                continue

            assert lineNumber != line.doppelganger, "line cannot be its own doppelganger"
            assert line.origin in "+-", "line with doppelganger must have origin"
            aheadOfDoppelganger = lineNumber < line.doppelganger

            if aheadOfDoppelganger:
                textA = line.text
                textB = self.lineData[line.doppelganger].text
                sm = difflib.SequenceMatcher(a=textA, b=textB)
                blocks = sm.get_matching_blocks()
                doppelgangerBlocksQueue.append(blocks)  # Set blocks aside for my doppelganger
            else:
                blocks = doppelgangerBlocksQueue.pop(0)  # Consume blocks set aside by my doppelganger

            if line.origin == "-":
                charFormat = DiffTextFormats.doppelgangerDelCF
            else:
                charFormat = DiffTextFormats.doppelgangerAddCF

            cursorPos = line.cursorStart
            oldBlockEnd = 0
            for blockStart, blockEnd in _invertMatchingBlocks(blocks, useA=aheadOfDoppelganger):
                assert blockStart >= oldBlockEnd

                # Advance Qt cursor (UTF-16!) to new doppelganger block
                cursorPos += qstringLength(line.text[oldBlockEnd: blockStart])
                cursor.setPosition(cursorPos, QTextCursor.MoveMode.MoveAnchor)

                # Move to end of doppelganger block and apply formatting
                cursorPos += qstringLength(line.text[blockStart: blockEnd])
                cursor.setPosition(cursorPos, QTextCursor.MoveMode.KeepAnchor)
                cursor.setCharFormat(charFormat)

                oldBlockEnd = blockEnd

        assert not doppelgangerBlocksQueue, "should've consumed all doppelganger matching blocks!"


def _invertMatchingBlocks(blockList: list[difflib.Match], useA: bool) -> Iterator[tuple[int, int]]:
    px = 0

    for block in blockList:
        x1 = block.a if useA else block.b
        x2 = x1 + block.size

        if px != x1:
            yield px, x1

        px = x2



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations  # TODO: Remove once we can drop support for Python <= 3.13

import enum

from gitfourchette.localization import *
from gitfourchette.qt import *
from gitfourchette.toolbox import *


class SearchProvider(QObject):
    class TermStatus(enum.IntEnum):
        Unknown = enum.auto()
        "Term hasn't been evaluated yet"

        Loading = enum.auto()
        "Asynchronous search in progress"

        Good = enum.auto()
        "Term is known to have some matches"

        Bad = enum.auto()
        "Term is known to have no matches"

    statusChanged = Signal(TermStatus)

    _term: str
    _badStem: str
    _status: TermStatus

    def __init__(self, parent):
        super().__init__(parent)
        self._term = ""
        self._badStem = ""
        self._status = SearchProvider.TermStatus.Unknown
        self._wantFilter = False
        self._frozen = False

    def freeze(self, frozen: bool):
        self._frozen = frozen
        if frozen:
            self.invalidate()

    def setFilterState(self, checked: bool):
        self._wantFilter = checked

    def term(self) -> str:
        return "" if self._frozen else self._term

    def setTerm(self, term: str):
        status1 = self._status

        with QSignalBlockerContext(self):  # Don't spam status updates
            term = term.strip().lower()
            self._term = term

            if term and self._badStem and self._badStem in term:
                # badStem is still in the search term: still bad
                assert self.isBad()
            else:
                self._badStem = ""
                self.setStatus(SearchProvider.TermStatus.Unknown)

            self._termChanged()

        if self._status != status1:
            self.statusChanged.emit(self._status)

    def _enshrineBadTerm(self):
        if not self._badStem:
            self._badStem = self._term
        self.setStatus(SearchProvider.TermStatus.Bad)

    def invalidate(self):
        self._badStem = ""
        self.setStatus(SearchProvider.TermStatus.Unknown)

    def status(self) -> SearchProvider.TermStatus:
        return self._status

    def setStatus(self, status: SearchProvider.TermStatus):
        if status != self._status:
            self._status = status
            self.statusChanged.emit(status)

    def isEmpty(self) -> bool:
        return not bool(self._term)

    def isBad(self) -> bool:
        return self._status == SearchProvider.TermStatus.Bad

    def isGoodAndNonEmpty(self) -> bool:
        return self._status == SearchProvider.TermStatus.Good and not self.isEmpty()

    # -------------------------------------------------------------------------
    # Override these

    def longTitle(self) -> str:
        return ""

    def shortTitle(self) -> str:
        return self.__class__.__name__

    def keyboardShortcut(self) -> str:
        return ""

    def canFilter(self) -> bool:
        return False

    def notFoundMessage(self) -> str:
        return _("No results")

    def _termChanged(self):
        pass

    def prime(self, forwardHint: bool):
        """
        Called by SearchBar to kick off a (possibly asynchronous) query on the
        search term, which is guaranteed to be non-empty with non-Bad status.
        This function can be a no-op; otherwise, it should typically change the
        current status to Good, Bad, or Loading.
        """

    def jump(self, forward: bool):
        """
        Jump to the next or previous occurrence.
        The search term status is guaranteed to be Good here.
        """
        raise NotImplementedError()

    def isCurrentMatch(self) -> bool:
        """
        Does the current selection match the current search term?
        The search term status is guaranteed to be Good here.
        """
        raise NotImplementedError()



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from contextlib import suppress

from gitfourchette import settings
from gitfourchette.application import GFApplication
from gitfourchette.exttools.usercommand import UserCommand
from gitfourchette.forms.searchbar import SearchBar
from gitfourchette.graphview.commitlogdelegate import CommitLogDelegate
from gitfourchette.graphview.commitlogfilter import CommitLogFilter
from gitfourchette.graphview.commitlogmodel import CommitLogModel, SpecialRow
from gitfourchette.graphview.commitfilesearch import CommitFileSearch
from gitfourchette.localization import *
from gitfourchette.nav import NavLocator, NavContext, NavFlags
from gitfourchette.porcelain import *
from gitfourchette.qt import *
from gitfourchette.repomodel import UC_FAKEID, GpgStatus, RepoModel
from gitfourchette.graphview.commitinfosearch import CommitInfoSearch
from gitfourchette.tasks import *
from gitfourchette.tasks.exporttasks import ExportABDiffAsPatch
from gitfourchette.toolbox import *


class GraphView(QListView):
    linkActivated = Signal(str)
    statusMessage = Signal(str)

    clModel: CommitLogModel
    clFilter: CommitLogFilter

    class SelectCommitError(KeyError):
        def __init__(self, oid: Oid, foundButHidden: bool, likelyTruncated: bool = False):
            super().__init__()
            self.oid = oid
            self.foundButHidden = foundButHidden
            self.likelyTruncated = likelyTruncated

        def __str__(self):
            if self.foundButHidden:
                m = _("This commit isn’t shown in the graph because it’s part of a hidden branch.")
            elif self.likelyTruncated:
                m = _("This commit isn’t shown in the graph because it isn’t part of the truncated commit history.")
            else:
                m = _("This commit isn’t shown in the graph.")
            return m

    def __init__(self, repoModel: RepoModel, parent):
        super().__init__(parent)

        self.repoModel = repoModel
        self.navLocator = NavLocator.Empty

        # Use tabular numbers (ISO dates look better with Inter, Cantarell, etc.)
        self.setFont(setFontFeature(self.font(), "tnum"))

        self.clModel = CommitLogModel(repoModel, self)
        self.clFilter = CommitLogFilter(repoModel, self)
        self.clFilter.setSourceModel(self.clModel)

        self.setModel(self.clFilter)

        self.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)

        # Massive perf boost when displaying/updating huge commit logs
        self.setUniformItemSizes(True)

        self.repoWidget = parent
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)  # prevents double-clicking to edit row text

        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self.onContextMenuRequested)

        # --------------
        # SearchBar

        infoSearch = CommitInfoSearch(self)
        fileSearch = CommitFileSearch(self)

        self.searchBar = SearchBar(self, infoSearch, fileSearch)
        self.searchBar.hide()

        # Invalidate the search when new commits trickle in at the top of the
        # graph. Connect the callback to clModel, not clFilter, so that:
        # a) CommitFileSearch gets to report that more results may be available
        #    in hidden branches;
        # b) Invalidating clFilter doesn't trash the search results, e.g. when
        #    toggling the 'Filter' checkbox.
        self.clModel.rowsInserted.connect(self.searchBar.reevaluateSearchTerm)

        # --------------

        self.clDelegate = CommitLogDelegate(self.repoModel, infoSearch, parent=self)
        self.setItemDelegate(self.clDelegate)

        GFApplication.instance().prefsChanged.connect(self.refreshPrefs)
        self.refreshPrefs(invalidateMetrics=False)

        # Shortcut keys
        makeWidgetShortcut(self, self.searchBar.hideOrBeep, "Escape")
        self.checkoutShortcut = makeWidgetShortcut(self, self.onReturnKey, "Return", "Enter")
        self.copyHashShortcut = makeWidgetShortcut(self, self.copyCommitHashToClipboard, QKeySequence.StandardKey.Copy)
        self.copyMessageShortcut = makeWidgetShortcut(self, self.copyCommitMessageToClipboard, "Ctrl+Shift+C")
        self.getInfoShortcut = makeWidgetShortcut(self, self.getInfoOnCurrentCommit, "Space")

    def mouseMoveEvent(self, event: QMouseEvent):
        """
        By default, ExtendedSelection lets the user select multiple items by
        holding down LMB and dragging. This event handler enforces single-item
        selection unless the user holds down Shift or Ctrl.
        """
        isLMB = bool(event.buttons() & Qt.MouseButton.LeftButton)
        isShift = bool(event.modifiers() & Qt.KeyboardModifier.ShiftModifier)
        isCtrl = bool(event.modifiers() & Qt.KeyboardModifier.ControlModifier)

        if isLMB and not isShift and not isCtrl:
            self.mousePressEvent(event)  # re-route event as if it were a click event
            self.scrollTo(self.indexAt(event.pos()))  # mousePressEvent won't scroll to the item on its own
        else:
            super().mouseMoveEvent(event)

    def mouseDoubleClickEvent(self, event: QMouseEvent):
        currentIndex = self.currentIndex()
        if (not currentIndex.isValid()
                or event.button() != Qt.MouseButton.LeftButton
                or event.modifiers() != Qt.KeyboardModifier.NoModifier):
            super().mouseDoubleClickEvent(event)
            return
        event.accept()
        rowKind = currentIndex.data(CommitLogModel.Role.SpecialRow)
        if rowKind == SpecialRow.UncommittedChanges:
            NewCommit.invoke(self)
        elif rowKind == SpecialRow.TruncatedHistory:
            self.linkActivated.emit(makeInternalLink("expandlog"))
        elif rowKind == SpecialRow.Commit:
            oid = self.currentCommitId
            CheckoutCommit.invoke(self, oid)

    def onReturnKey(self):
        oid = self.currentCommitId
        isValidCommit = oid and oid != UC_FAKEID
        if isValidCommit:
            CheckoutCommit.invoke(self, oid)
        else:
            NewCommit.invoke(self)

    @property
    def currentRowKind(self) -> SpecialRow:
        # TODO: The only remaining uses of this function are in unit tests. Remove?
        if self.navLocator.context == NavContext.COMMITTED:
            return SpecialRow.Commit
        elif self.navLocator.context.isWorkdir():
            return SpecialRow.UncommittedChanges
        elif self.navLocator.context == NavContext.SPECIAL:
            return SpecialRow.fromString(self.navLocator.path)
        else:
            return SpecialRow.Invalid

    @property
    def currentCommitId(self) -> Oid | None:
        # TODO: If pygit2 had Oid.__bool__() which returned True if the hash isn't NULL_OID,
        #       we wouldn't have to return None for compatibility with existing code
        #       (pygit2 1.18.0+ has this now)
        if self.navLocator.context == NavContext.COMMITTED:
            return self.navLocator.commit
        else:
            return None

    def getInfoOnCurrentCommit(self):
        oid = self.currentCommitId
        if not oid:
            return
        withDebugInfo = QGuiApplication.keyboardModifiers() & (Qt.KeyboardModifier.AltModifier | Qt.KeyboardModifier.ShiftModifier)
        GetCommitInfo.invoke(self, oid, withDebugInfo)

    def copyCommitHashToClipboard(self):
        oid = self.currentCommitId
        if not oid:  # uncommitted changes
            return
        text = str(oid)
        QApplication.clipboard().setText(text)
        self.statusMessage.emit(clipboardStatusMessage(text))

    def copyCommitMessageToClipboard(self):
        oid = self.currentCommitId
        if not oid:  # uncommitted changes
            return
        commit = self.repoModel.repo[oid].peel(Commit)
        text = commit.message.rstrip()
        QApplication.clipboard().setText(text)
        self.statusMessage.emit(clipboardStatusMessage(text))

    def selectionChanged(self, selected: QItemSelection, deselected: QItemSelection):
        # do standard callback, such as scrolling the viewport if reaching the edges, etc.
        super().selectionChanged(selected, deselected)

        # Don't bother with the jump if our signals are blocked
        if self.signalsBlocked():
            return

        locator = self.locatorFromSelection()
        if not locator:
            return

        locator = locator.withExtraFlags(NavFlags.BypassCommitSelect)
        Jump.invoke(self, locator)

    def locatorFromSelection(self):
        # Warning: selection not sorted!
        indexes = self.selectedIndexes()
        if not indexes:
            return NavLocator.Empty

        # Get the "last" selected index.
        # Note that currentIndex() may return an index outside the selection!
        current = self.currentIndex()
        if not current.isValid() or current not in indexes:
            current = indexes[-1]
        assert current.isValid()

        special = current.data(CommitLogModel.Role.SpecialRow)
        if special == SpecialRow.UncommittedChanges:
            locator = NavLocator.inWorkdir()
        elif special == SpecialRow.Commit:
            oid = current.data(CommitLogModel.Role.Oid)
            locator = NavLocator.inCommit(oid)
        else:
            locator = NavLocator.inSpecial(special)

        # Handle multiple selected rows
        if len(indexes) > 1:
            originalCommit = locator.commit

            # A/B diffing only allowed if exactly two commits are selected
            twoSelected = len(indexes) == 2

            # When initiating an A/B diff, ensure the "B" side comes out on top
            # regardless of the order in which the selection was made.
            if twoSelected:
                indexes.sort(key=lambda i: i.row(), reverse=True)

            # Get commit oids
            oids = tuple(i.data(CommitLogModel.Role.Oid) for i in indexes)

            if not twoSelected:
                locator = NavLocator.inSpecial(SpecialRow.TooManyRowsSelected)
            elif any(o in (UC_FAKEID, None) for o in oids):
                # Both oids must be valid commit oids
                locator = NavLocator.inSpecial(SpecialRow.CannotCompareRows)

            # Keep track of selected/current rows
            locator = locator.replace(selectedCommits=oids, commit=originalCommit)

        return locator

    def selectRowForLocator(self, locator: NavLocator):
        # Keep scroll position if we're re-selecting the same row
        same = locator.isSimilarEnoughTo(self.navLocator, considerPaths=False)
        vpos = self.verticalScrollBar().sliderPosition()

        # Save locator as current
        self.navLocator = locator

        # Update A/B commits in model
        commitDiffAB = locator.commitDiffAB()
        if commitDiffAB != self.clModel.commitDiffAB:
            self.clModel.commitDiffAB = commitDiffAB
            self.viewport().update()  # Redraw A/B icons (or lack thereof)

        # Bail if this call originates from a user click (selection already correct)
        if locator.hasFlags(NavFlags.BypassCommitSelect):
            return

        # Get index for main selected row
        filterIndex = self.getFilterIndexForLocator(locator)

        # Update selection model
        with QSignalBlockerContext(self):  # DON'T emit jump signal
            sm = self.selectionModel()
            sm.clear()

            # Multiple selected commits
            for sc in locator.selectedCommits:
                fi = self.getFilterIndexForCommit(sc)
                sm.select(fi, QItemSelectionModel.SelectionFlag.Select)

            # Set current index (i.e. keyboard focus) on main selected row.
            # This automatically scrolls the row into view.
            sm.setCurrentIndex(filterIndex, QItemSelectionModel.SelectionFlag.Select)

        if same:
            self.verticalScrollBar().setSliderPosition(vpos)

    def getFilterIndexForLocator(self, locator: NavLocator) -> QModelIndex:
        if locator.context == NavContext.COMMITTED:
            index = self.getFilterIndexForCommit(locator.commit)
            assert index.data(CommitLogModel.Role.SpecialRow) == SpecialRow.Commit
            return index

        if locator.context.isWorkdir():
            index = self.clFilter.index(0, 0)
            if index.data(CommitLogModel.Role.SpecialRow) != SpecialRow.UncommittedChanges:
                # Workdir row is being filtered out
                raise self.SelectCommitError(None, True, False)
            return index

        if locator.context == NavContext.SPECIAL:
            special = SpecialRow.fromString(locator.path)

            # Multi-selection error pages: use main commit row
            if special in (SpecialRow.CannotCompareRows, SpecialRow.TooManyRowsSelected):
                return self.getFilterIndexForCommit(locator.commit)

            # Last row
            if self.clModel._extraRow != special:
                raise GraphView.SelectCommitError(None, False, False)
            index = self.clFilter.index(self.clFilter.rowCount()-1, 0)
            assert index.data(CommitLogModel.Role.SpecialRow) == special
            return index

        raise NotImplementedError(f"unsupported locator context {locator.context}")

    def getFilterIndexForCommit(self, oid: Oid) -> QModelIndex:
        try:
            rawIndex = self.repoModel.graph.getCommitRow(oid)
        except KeyError as exc:
            raise GraphView.SelectCommitError(oid, foundButHidden=False, likelyTruncated=self.repoModel.truncatedHistory) from exc

        newSourceIndex = self.clModel.index(rawIndex, 0)
        newFilterIndex = self.clFilter.mapFromSource(newSourceIndex)

        if not newFilterIndex.isValid():
            raise GraphView.SelectCommitError(oid, foundButHidden=True)

        return newFilterIndex

    def isLocatorVisible(self, locator: NavLocator) -> bool:
        try:
            self.getFilterIndexForLocator(locator)
            return True
        except GraphView.SelectCommitError:
            return False

    def scrollToRowForLocator(self, locator: NavLocator, scrollHint=QAbstractItemView.ScrollHint.EnsureVisible):
        with suppress(GraphView.SelectCommitError):
            filterIndex = self.getFilterIndexForLocator(locator)
            self.scrollTo(filterIndex, scrollHint)

    def repaintCommit(self, oid: Oid):
        with suppress(GraphView.SelectCommitError):
            filterIndex = self.getFilterIndexForCommit(oid)
            self.update(filterIndex)

    def refreshPrefs(self, invalidateMetrics=True):
        self.setVerticalScrollMode(settings.prefs.listViewScrollMode)
        self.setAlternatingRowColors(settings.prefs.alternatingRowColors)

        # Force redraw to reflect changes in row height, flattening, date format, etc.
        if invalidateMetrics:
            self.clDelegate.invalidateMetrics()
            self.model().layoutChanged.emit()

    # -------------------------------------------------------------------------
    # Context menus

    def onContextMenuRequested(self, point: QPoint):
        actions = None
        locator = self.navLocator

        if locator.context.isWorkdir():
            actions = self._contextMenuActionsUncommittedChanges()

        elif locator.context == NavContext.COMMITTED:
            if locator.commitDiffAB():
                actions = self._contextMenuActions2Commits(locator)
            else:
                actions = self._contextMenuActions1Commit()

        elif locator.context == NavContext.SPECIAL:
            special = SpecialRow.fromString(locator.path)
            if special == SpecialRow.TruncatedHistory:
                actions = self._contextMenuActionsTruncatedHistory()

        # Fall back to no-op menu
        if actions is None:
            actions = [
                ActionDef(_("No actions available for this selection"), enabled=False)
            ]

        menu = ActionDef.makeQMenu(self, actions)
        menu.setObjectName("GraphViewCM")
        menu.aboutToHide.connect(menu.deleteLater)
        menu.popup(self.mapToGlobal(point))

    def _contextMenuActionsUncommittedChanges(self):
        mainWindow = GFApplication.instance().mainWindow

        actions = [
            TaskBook.action(self, NewCommit, accel="C"),
            TaskBook.action(self, AmendCommit, accel="A"),
            ActionDef.SEPARATOR,
            TaskBook.action(self, NewStash, accel="S"),
            TaskBook.action(self, ExportWorkdirAsPatch, accel="X"),
        ]

        if self.repoModel.prefs.hasDraftCommit():
            actions.extend([
                ActionDef.SEPARATOR,
                ActionDef(_("Clear Draft Message"), self.repoModel.prefs.clearDraftCommit),
            ])

        actions.extend(mainWindow.contextualUserCommands(UserCommand.Token.Workdir))
        return actions

    def _contextMenuActionsTruncatedHistory(self):
        locale = self.locale()
        expandSome = makeInternalLink("expandlog")
        expandAll = makeInternalLink("expandlog", n=str(0))
        changePref = makeInternalLink("prefs", "maxCommits")
        actions = [
            ActionDef(_("Load up to {0} commits", locale.toString(self.repoModel.nextTruncationThreshold)),
                      lambda: self.linkActivated.emit(expandSome)),

            ActionDef(_("Load full commit history"),
                      lambda: self.linkActivated.emit(expandAll)),

            ActionDef(_("Change threshold setting"),
                      lambda: self.linkActivated.emit(changePref)),
        ]
        return actions

    def _contextMenuActions1Commit(self):
        mainWindow = GFApplication.instance().mainWindow
        repoModel = self.repoModel
        repo = repoModel.repo
        oid = self.currentCommitId
        assert oid is not None

        myRef = lquo(repoModel.homeBranch) if repoModel.homeBranch else "HEAD"

        # Figure out a nice ref name to initiate a merge, or fall back to commit id
        mergeWhat: str | Oid
        try:
            refsHere = repoModel.refsAt[oid]
            mergeWhat = next(ref for ref in refsHere if ref.startswith((RefPrefix.HEADS, RefPrefix.REMOTES)))
        except (KeyError, StopIteration):
            mergeWhat = oid

        checkoutAction = TaskBook.action(self, CheckoutCommit, _("&Check Out…"), taskArgs=oid)
        checkoutAction.shortcuts = self.checkoutShortcut.key()

        gpgLookAtCommit = repo.peel_commit(oid)
        gpgStatus, _gpgKeyInfo = repoModel.getCachedGpgStatus(gpgLookAtCommit)
        gpgIcon = gpgStatus.iconName()

        mounts = GFApplication.instance().mountManager
        mountCaption = _("&Mount Commit As Folder")
        if not mounts.supportsMounting():
            mountActions = []
        elif mounts.isMounted(oid):
            mountActions = [ActionDef(mountCaption, icon="git-mount", submenu=mounts.makeMenuItemsForMount(oid, self))]
        else:
            mountActions = [ActionDef(mountCaption, icon="git-mount", callback=lambda: mounts.mount(repo.workdir, oid))]

        actions = [
            TaskBook.action(self, NewBranchFromCommit, _("New &Branch Here…"), taskArgs=oid),
            TaskBook.action(self, NewTag, _("&Tag This Commit…"), taskArgs=oid),
            ActionDef.SEPARATOR,
            checkoutAction,
            TaskBook.action(self, MergeBranch, _("&Merge into {0}…", myRef), taskArgs=(mergeWhat,)),
            TaskBook.action(self, ResetHead, _("&Reset {0} to Here…", myRef), taskArgs=oid),
            ActionDef.SEPARATOR,
            TaskBook.action(self, CherrypickCommit, _("Cherry &Pick…"), taskArgs=oid),
            TaskBook.action(self, RevertCommit, _("Re&vert…"), taskArgs=oid),
            TaskBook.action(self, ExportCommitAsPatch, _("E&xport As Patch…"), taskArgs=oid),
            ActionDef.SEPARATOR,
            ActionDef(_("Copy Commit &Hash"), self.copyCommitHashToClipboard, shortcuts=self.copyHashShortcut.key()),
            ActionDef(_("Copy Commit M&essage"), self.copyCommitMessageToClipboard, shortcuts=self.copyMessageShortcut.key()),
            TaskBook.action(self, VerifyGpgSignature, taskArgs=oid, enabled=gpgStatus != GpgStatus.Unsigned, icon=gpgIcon, accel="G"),
            *mountActions,
            ActionDef(_("Get &Info…"), self.getInfoOnCurrentCommit, "SP_MessageBoxInformation", shortcuts=self.getInfoShortcut.key()),
            *mainWindow.contextualUserCommands(UserCommand.Token.Commit),
        ]
        return actions

    def _contextMenuActions2Commits(self, locator: NavLocator):
        diffAB = locator.commitDiffAB()

        actions = [
            ActionDef(_("&Swap A/B"),
                      lambda: Jump.invoke(self, locator.swapABCommits())),

            ActionDef(_("E&xport A/B Diff As Patch…"),
                      lambda: ExportABDiffAsPatch.invoke(self, diffAB))
        ]
        return actions
